/*
 *  Copyright (c) 2000 LSI Logic Corporation.
 *
 *
 *           Name:  MPI_CNFG.H
 *          Title:  MPI Config message, structures, and Pages
 *  Creation Date:  July 27, 2000
 *
 *    MPI Version:  01.00.05
 *
 *  Version History
 *  ---------------
 *   File Version:  %version: %
 *   Date Updated:  %date_modified: %
 *     Programmer:  %created_by: %
 *
 *
 *  Date      Version   Description
 *  --------  --------  ------------------------------------------------------
 *  05-08-00  00.10.01  Original release for 0.10 spec dated 4/26/2000.
 *  06-06-00  01.00.01  Update version number for 1.0 release.
 *  06-08-00  01.00.02  Added _PAGEVERSION definitions for all pages.
 *                      Added FcPhLowestVersion, FcPhHighestVersion, Reserved2
 *                      fields to FC_DEVICE_0 page, updated the page version.
 *                      Changed _FREE_RUNNING_CLOCK to _PACING_TRANSFERS in
 *                      SCSI_PORT_0, SCSI_DEVICE_0 and SCSI_DEVICE_1 pages
 *                      and updated the page versions.
 *                      Added _RESPONSE_ID_MASK definition to SCSI_PORT_1
 *                      page and updated the page version.
 *                      Added Information field and _INFO_PARAMS_NEGOTIATED 
 *                      definitionto SCSI_DEVICE_0 page.
 *  06-22-00  01.00.03  Removed batch controls from LAN_0 page and updated the
 *                      page version.
 *                      Added BucketsRemaining to LAN_1 page, redefined the
 *                      state values, and updated the page version.
 *                      Revised bus width definitions in SCSI_PORT_0,
 *                      SCSI_DEVICE_0 and SCSI_DEVICE_1 pages. 
 *  06-30-00  01.00.04  Added MaxReplySize to LAN_1 page and updated the page
 *                      version.
 *                      Moved FC_DEVICE_0 PageAddress description to spec.
 *  07-27-00  01.00.05  Corrected the SubsystemVendorID and SubsystemID field 
 *                      widths in IOC_0 page and updated the page version.
 *
 *  --------------------------------------------------------------------------
 */

#ifndef MPI_CNFG_H
#define MPI_CNFG_H


/*****************************************************************************
*
*       C o n f i g    M e s s a g e    a n d    S t r u c t u r e s
*
*****************************************************************************/

typedef struct _CONFIG_PAGE_HEADER
{
    U8                      PageVersion;
    U8                      PageLength;
    U8                      PageNumber;
    U8                      PageType;
} CONFIG_PAGE_HEADER, *PTR_CONFIG_PAGE_HEADER,
  ConfigPageHeader_t, *pConfigPageHeader_t;

typedef union _CONFIG_PAGE_HEADER_UNION
{
   ConfigPageHeader_t  Struct;
   U8                  Bytes[4];
   U16                 Word16[2];
   U32                 Word32;
} ConfigPageHeaderUnion, *pConfigPageHeaderUnion,
  CONFIG_PAGE_HEADER_UNION, *PTR_CONFIG_PAGE_HEADER_UNION;


/****************************************************************************/
/*  PageType field values                                                   */
/****************************************************************************/

#define MPI_CONFIG_PAGEATTR_READ_ONLY               (0x00)
#define MPI_CONFIG_PAGEATTR_CHANGEABLE              (0x10)
#define MPI_CONFIG_PAGEATTR_PERSISTENT              (0x20)
#define MPI_CONFIG_PAGEATTR_MASK                    (0xF0)

#define MPI_CONFIG_PAGETYPE_IO_UNIT                 (0x00)
#define MPI_CONFIG_PAGETYPE_IOC                     (0x01)
#define MPI_CONFIG_PAGETYPE_BIOS                    (0x02)
#define MPI_CONFIG_PAGETYPE_SCSI_PORT               (0x03)
#define MPI_CONFIG_PAGETYPE_SCSI_DEVICE             (0x04)
#define MPI_CONFIG_PAGETYPE_FC_PORT                 (0x05)
#define MPI_CONFIG_PAGETYPE_FC_DEVICE               (0x06)
#define MPI_CONFIG_PAGETYPE_LAN                     (0x07)
#define MPI_CONFIG_PAGETYPE_MASK                    (0x0F)

#define MPI_CONFIG_TYPENUM_MASK                     (0x0FFF)


/****************************************************************************/
/*  Config Message                                                          */
/****************************************************************************/
typedef struct _MSG_CONFIG
{
    U8                      Action;
    U8                      Reserved;
    U8                      ChainOffset;
    U8                      Function;
    U8                      Reserved1[3];
    U8                      MsgFlags;
    U32                     MsgContext;
    U8                      Reserved2[8];
    CONFIG_PAGE_HEADER      Header;
    U32                     PageAddress;
    SGE_IO_UNION            PageBufferSGE;
} MSG_CONFIG, *PTR_MSG_CONFIG,
  Config_t, *pConfig_t;


/****************************************************************************/
/*  Action field values                                                     */
/****************************************************************************/
#define MPI_CONFIG_ACTION_PAGE_HEADER               (0x00)
#define MPI_CONFIG_ACTION_PAGE_READ                 (0x01)
#define MPI_CONFIG_ACTION_PAGE_WRITE                (0x02)
#define MPI_CONFIG_ACTION_PAGE_DEFAULT              (0x03)
#define MPI_CONFIG_ACTION_PAGE_WRITE_COMMIT         (0x04)

/* Config Reply Message */

typedef struct _MSG_CONFIG_REPLY
{
    U8                      Action;
    U8                      Reserved;
    U8                      MsgLength;
    U8                      Function;
    U8                      Reserved1[3];
    U8                      MsgFlags;
    U32                     MsgContext;
    U8                      Reserved2[2];
    U16                     IOCStatus;
    U32                     IOCLogInfo;
    CONFIG_PAGE_HEADER      Header;
} MSG_CONFIG_REPLY, *PTR_MSG_CONFIG_REPLY,
  ConfigReply_t, *pConfigReply_t;




/*****************************************************************************
*
*               C o n f i g u r a t i o n    P a g e s
*
*****************************************************************************/

/****************************************************************************/
/*  IO Unit Config Pages                                                    */
/****************************************************************************/

typedef struct CONFIG_PAGE_IO_UNIT_0
{
    CONFIG_PAGE_HEADER      Header;
    U64                     UniqueValue;
} CONFIG_PAGE_IO_UNIT_0, *PTR_CONFIG_PAGE_IO_UNIT_0,
  IOUnitPage0_t, *pIOUnitPage0_t;

#define MPI_IOUNITPAGE0_PAGEVERSION                 (0x00)

typedef struct _CONFIG_PAGE_IO_UNIT_1
{
    CONFIG_PAGE_HEADER      Header;
    U32                     Flags;
} CONFIG_PAGE_IO_UNIT_1, *PTR_CONFIG_PAGE_IO_UNIT_1,
  IOUnitPage1_t, *pIOUnitPage1_t;

#define MPI_IOUNITPAGE1_PAGEVERSION                 (0x00)

#define MPI_IOUNITPAGE1_MULTI_FUNCTION              (0x00000000)
#define MPI_IOUNITPAGE1_SINGLE_FUNCTION             (0x00000001)
#define MPI_IOUNITPAGE1_MULTI_PATHING               (0x00000002)
#define MPI_IOUNITPAGE1_SINGLE_PATHING              (0x00000000)

#define MPI_IOUNITPAGE1_FORCE_32                    (0x00000080)

/****************************************************************************/
/*  IOC Config Pages                                                        */
/****************************************************************************/

typedef struct _CONFIG_PAGE_IOC_0
{
    CONFIG_PAGE_HEADER      Header;
    U32                     TotalNVStore;
    U32                     FreeNVStore;
    U16                     VendorID;
    U16                     DeviceID;
    U8                      RevisionID;
    U8                      Reserved[3];
    U32                     ClassCode;
    U16                     SubsystemVendorID;
    U16                     SubsystemID;
} CONFIG_PAGE_IOC_0, *PTR_CONFIG_PAGE_IOC_0,
  IOCPage0_t, *pIOCPage0_t;

#define MPI_IOCPAGE0_PAGEVERSION                    (0x01)

typedef struct _CONFIG_PAGE_IOC_1
{
    CONFIG_PAGE_HEADER      Header;
    U32                     Flags;
    U32                     CoalescingTimeout;
    U8                      CoalescingDepth;
    U8                      Reserved[3];
} CONFIG_PAGE_IOC_1, *PTR_CONFIG_PAGE_IOC_1,
  IOCPage1_t, *pIOCPage1_t;

#define MPI_IOCPAGE1_PAGEVERSION                    (0x00)

#define MPI_IOCPAGE1_REPLY_COALESCING               (0x00000001)


/****************************************************************************/
/*  SCSI Port Config Pages                                                  */
/****************************************************************************/

typedef struct _CONFIG_PAGE_SCSI_PORT_0
{
    CONFIG_PAGE_HEADER      Header;
    U32                     Capabilities;
    U32                     PhysicalInterface;
} CONFIG_PAGE_SCSI_PORT_0, *PTR_CONFIG_PAGE_SCSI_PORT_0,
  SCSIPortPage0_t, *pSCSIPortPage0_t;

#define MPI_SCSIPORTPAGE0_PAGEVERSION               (0x01)

/* Capabilities */
#define MPI_SCSIPORTPAGE0_CAP_IU                    (0x00000001)
#define MPI_SCSIPORTPAGE0_CAP_DT                    (0x00000002)
#define MPI_SCSIPORTPAGE0_CAP_QAS                   (0x00000004)
#define MPI_SCSIPORTPAGE0_CAP_PACING_TRANSFERS      (0x00000008)
#define MPI_SCSIPORTPAGE0_CAP_MIN_SYNC_PERIOD_MASK  (0x0000FF00)
#define MPI_SCSIPORTPAGE0_CAP_MAX_SYNC_OFFSET_MASK  (0x00FF0000)
#define MPI_SCSIPORTPAGE0_CAP_WIDE                  (0x20000000)
#define MPI_SCSIPORTPAGE0_CAP_AIP                   (0x80000000)

/* Physical Interface   */
#define MPI_SCSIPORTPAGE0_PHY_SIGNAL_TYPE_MASK      (0x00000007)
#define MPI_SCSIPORTPAGE0_PHY_SIGNAL_HVD            (0x01)
#define MPI_SCSIPORTPAGE0_PHY_SIGNAL_SE             (0x02)
#define MPI_SCSIPORTPAGE0_PHY_SIGNAL_LVD            (0x03)

typedef struct _CONFIG_PAGE_SCSI_PORT_1
{
    CONFIG_PAGE_HEADER      Header;
    U32                     Configuration;
} CONFIG_PAGE_SCSI_PORT_1, *PTR_CONFIG_PAGE_SCSI_PORT_1,
  SCSIPortPage1_t, *pSCSIPortPage1_t;

#define MPI_SCSIPORTPAGE1_PAGEVERSION               (0x01)

#define MPI_SCSIPORTPAGE1_CFG_PORT_SCSI_ID_MASK     (0x000000FF)
#define MPI_SCSIPORTPAGE1_CFG_DEFER_NEGOTIATIONS    (0x00000100)
#define MPI_SCSIPORTPAGE1_CFG_PORT_RESPONSE_ID_MASK (0xFFFF0000)


/****************************************************************************/
/*  SCSI Target Device Config Pages                                         */
/****************************************************************************/

typedef struct _CONFIG_PAGE_SCSI_DEVICE_0
{
    CONFIG_PAGE_HEADER      Header;
    U32                     NegotiatedParameters;
    U32                     Information;
} CONFIG_PAGE_SCSI_DEVICE_0, *PTR_CONFIG_PAGE_SCSI_DEVICE_0,
  SCSIDevicePage0_t, *pSCSIDevicePage0_t;

#define MPI_SCSIDEVPAGE0_PAGEVERSION             (0x01)

#define MPI_SCSIDEVPAGE0_NP_IU                   (0x00000001)
#define MPI_SCSIDEVPAGE0_NP_DT                   (0x00000002)
#define MPI_SCSIDEVPAGE0_NP_QAS                  (0x00000004)
#define MPI_SCSIDEVPAGE0_NP_PACING_TRANSFERS     (0x00000008)
#define MPI_SCSIDEVPAGE0_NP_NEG_SYNC_PERIOD_MASK (0x0000FF00)
#define MPI_SCSIDEVPAGE0_NP_NEG_SYNC_OFFSET_MASK (0x00FF0000)
#define MPI_SCSIDEVPAGE0_NP_WIDE                 (0x20000000)
#define MPI_SCSIDEVPAGE0_NP_AIP                  (0x80000000)

#define MPI_SCSIDEVPAGE0_INFO_PARAMS_NEGOTIATED  (0x00000001)

typedef struct _CONFIG_PAGE_SCSI_DEVICE_1
{
    CONFIG_PAGE_HEADER      Header;
    U32                     RequestedParameters;
    U32                     DomainValidation;
    U32                     Configuration;
} CONFIG_PAGE_SCSI_DEVICE_1, *PTR_CONFIG_PAGE_SCSI_DEVICE_1,
  SCSIDevicePage1_t, *pSCSIDevicePage1_t;

#define MPI_SCSIDEVPAGE1_PAGEVERSION                 (0x01)

#define MPI_SCSIDEVPAGE1_RP_IU                       (0x00000001)
#define MPI_SCSIDEVPAGE1_RP_DT                       (0x00000002)
#define MPI_SCSIDEVPAGE1_RP_QAS                      (0x00000004)
#define MPI_SCSIDEVPAGE1_RP_PACING_TRANSFERS         (0x00000008)
#define MPI_SCSIDEVPAGE1_RP_MIN_SYNC_PERIOD_MASK     (0x0000FF00)
#define MPI_SCSIDEVPAGE1_RP_MAX_SYNC_OFFSET_MASK     (0x00FF0000)
#define MPI_SCSIDEVPAGE1_RP_WIDE                     (0x20000000)
#define MPI_SCSIDEVPAGE1_RP_AIP                      (0x80000000)

#define MPI_SCSIDEVPAGE1_DV_LVD_DRIVE_STRENGTH_MASK  (0x00000003)
#define MPI_SCSIDEVPAGE1_DV_SE_SLEW_RATE_MASK        (0x00000300)

#define MPI_SCSIDEVPAGE1_CONF_PPR_ALLOWED            (0x00000001)

/****************************************************************************/
/*  FC Port Config Pages                                                    */
/****************************************************************************/

typedef struct _CONFIG_PAGE_FC_PORT_0
{
    CONFIG_PAGE_HEADER      Header;
    U32                     Flags;
    U8                      MPIPortNumber;
    U8                      Reserved[3];
    U32                     PortIdentifier;
    U64                     WWNN;
    U64                     WWPN;
} CONFIG_PAGE_FC_PORT_0, *PTR_CONFIG_PAGE_FC_PORT_0,
  FCPortPage0_t, *pFCPortPage0_t;

#define MPI_FCPORTPAGE0_PAGEVERSION                 (0x00)

#define MPI_FCPORTPAGE0_FLAGS_PROT_MASK             (0x0000000F)
#define MPI_FCPORTPAGE0_FLAGS_PROT_FCP_INIT         (MPI_PORTFACTS_PROTOCOL_INITIATOR)
#define MPI_FCPORTPAGE0_FLAGS_PROT_FCP_TARG         (MPI_PORTFACTS_PROTOCOL_TARGET)
#define MPI_FCPORTPAGE0_FLAGS_PROT_LAN              (MPI_PORTFACTS_PROTOCOL_LAN)
#define MPI_FCPORTPAGE0_FLAGS_PROT_LOGBUSADDR       (MPI_PORTFACTS_PROTOCOL_LOGBUSADDR)

#define MPI_FCPORTPAGE0_FLAGS_ATTACH_TYPE_MASK      (0x00000F00)
#define MPI_FCPORTPAGE0_FLAGS_ATTACH_PUBLIC_LOOP    (0x00000800)
#define MPI_FCPORTPAGE0_FLAGS_ATTACH_FABRIC_DIRECT  (0x00000400)
#define MPI_FCPORTPAGE0_FLAGS_ATTACH_PRIVATE_LOOP   (0x00000200)
#define MPI_FCPORTPAGE0_FLAGS_ATTACH_POINT_TO_POINT (0x00000100)
#define MPI_FCPORTPAGE0_FLAGS_ATTACH_NO_INIT        (0x00000000)


typedef struct _CONFIG_PAGE_FC_PORT_1
{
    CONFIG_PAGE_HEADER      Header;
    U32                     Flags;
    U64                     NoSEEPROMWWNN;
    U64                     NoSEEPROMWWPN;
    U8                      HardALPA;
    U8                      Reserved[3];
} CONFIG_PAGE_FC_PORT_1, *PTR_CONFIG_PAGE_FC_PORT_1,
  FCPortPage1_t, *pFCPortPage1_t;

#define MPI_FCPORTPAGE1_PAGEVERSION                 (0x00)

#define MPI_FCPORTPAGE1_FLAGS_SORT_BY_DID           (0x00000001)
#define MPI_FCPORTPAGE1_FLAGS_SORT_BY_WWN           (0x00000000)

/* Flags used for programming protocol modes in NVStore */
#define MPI_FCPORTPAGE1_FLAGS_PROT_MASK             (0xF0000000)
#define MPI_FCPORTPAGE1_FLAGS_PROT_SHIFT            (28)
#define MPI_FCPORTPAGE1_FLAGS_PROT_FCP_INIT         ((U32)MPI_PORTFACTS_PROTOCOL_INITIATOR << MPI_FCPORTPAGE1_FLAGS_PROT_SHIFT)
#define MPI_FCPORTPAGE1_FLAGS_PROT_FCP_TARG         ((U32)MPI_PORTFACTS_PROTOCOL_TARGET << MPI_FCPORTPAGE1_FLAGS_PROT_SHIFT)
#define MPI_FCPORTPAGE1_FLAGS_PROT_LAN              ((U32)MPI_PORTFACTS_PROTOCOL_LAN << MPI_FCPORTPAGE1_FLAGS_PROT_SHIFT)
#define MPI_FCPORTPAGE1_FLAGS_PROT_LOGBUSADDR       ((U32)MPI_PORTFACTS_PROTOCOL_LOGBUSADDR << MPI_FCPORTPAGE1_FLAGS_PROT_SHIFT)

#define MPI_FCPORTPAGE1_HARD_ALPA_NOT_USED          (0xFF)


typedef struct _CONFIG_PAGE_FC_PORT_2
{
    CONFIG_PAGE_HEADER      Header;
    U8                      NumberActive;
    U8                      ALPA[126];
    U8                      Reserved;
} CONFIG_PAGE_FC_PORT_2, *PTR_CONFIG_PAGE_FC_PORT_2,
  FCPortPage2_t, *pFCPortPage2_t;

#define MPI_FCPORTPAGE2_PAGEVERSION                 (0x00)

typedef struct _FC_PORT_PERSISTENT
{
    U64                     WWNN;
    U64                     WWPN;
    U8                      TargetID;
    U8                      Bus;
    U16                     Flags;
} FC_PORT_PERSISTENT, *PTR__FC_PORT_PERSISTENT,
  PersistentData_t, *pPersistentData_t;

#define MPI_PERSISTENT_FLAGS_SHIFT                  (16)
#define MPI_PERSISTENT_FLAGS_ENTRY_VALID            (0x0001)

typedef struct _CONFIG_PAGE_FC_PORT_3
{
    CONFIG_PAGE_HEADER      Header;
    FC_PORT_PERSISTENT      Entry[1];
} CONFIG_PAGE_FC_PORT_3, *PTR_CONFIG_PAGE_FC_PORT_3,
  FCPortPage3_t, *pFCPortPage3_t;

#define MPI_FCPORTPAGE3_PAGEVERSION                 (0x00)


/****************************************************************************/
/*  FC Device Config Pages                                                  */
/****************************************************************************/

typedef struct _CONFIG_PAGE_FC_DEVICE_0
{
    CONFIG_PAGE_HEADER      Header;
    U64                     WWNN;
    U64                     WWPN;
    U32                     PortIdentifier;
    U8                      Protocol;
    U8                      Reserved;
    U16                     BBCredit;
    U16                     MaxRxFrameSize;
    U8                      Reserved1;
    U8                      PortNumber;
    U8                      FcPhLowestVersion;
    U8                      FcPhHighestVersion;
    U16                     Reserved2;
} CONFIG_PAGE_FC_DEVICE_0, *PTR_CONFIG_PAGE_FC_DEVICE_0,
  FCDevicePage0_t, *pFCDevicePage0_t;

#define MPI_FC_DEVICE_PAGE_0_PAGEVERSION            (0x01)

#define MPI_FC_DEVICE_PAGE_0_PROT_IP                (0x01)
#define MPI_FC_DEVICE_PAGE_0_PROT_FCP_TARGET        (0x02)
#define MPI_FC_DEVICE_PAGE_0_PROT_FCP_INITIATOR     (0x04)

#define MPI_FC_DEVICE_PAGE_0_PGAD_PORT_MASK         (0xF0000000)
#define MPI_FC_DEVICE_PAGE_0_PGAD_FORM_MASK         (0x0F000000)
#define MPI_FC_DEVICE_PAGE_0_PGAD_FORM_NEXT_DID     (0x00000000)
#define MPI_FC_DEVICE_PAGE_0_PGAD_FORM_BUS_TID      (0x01000000)
#define MPI_FC_DEVICE_PAGE_0_PGAD_DID_MASK          (0x00FFFFFF)
#define MPI_FC_DEVICE_PAGE_0_PGAD_BUS_MASK          (0x0000FF00)
#define MPI_FC_DEVICE_PAGE_0_PGAD_BUS_SHIFT         (8)
#define MPI_FC_DEVICE_PAGE_0_PGAD_TID_MASK          (0x000000FF)


/****************************************************************************/
/* LAN Config Pages                                                         */
/****************************************************************************/

typedef struct _CONFIG_PAGE_LAN_0
{
    ConfigPageHeader_t      Header;
    U16                     TxRxModes;
    U16                     Reserved;
    U32                     PacketPrePad;
} CONFIG_PAGE_LAN_0, *PTR_CONFIG_PAGE_LAN_0,
  LANPage0_t, *pLANPage0_t;

#define MPI_LAN_PAGE_0_PAGEVERSION                  (0x01)

#define MPI_LAN_PAGE_0_RETURN_LOOPBACK              (0x0000)
#define MPI_LAN_PAGE_0_SUPPRESS_LOOPBACK            (0x0001)
#define MPI_LAN_PAGE_0_LOOPBACK_MASK                (0x0001)

typedef struct _CONFIG_PAGE_LAN_1
{
    ConfigPageHeader_t      Header;
    U16                     Reserved;
    U8                      CurrentDeviceState;
    U8                      Reserved1;
    U32                     MinPacketSize;
    U32                     MaxPacketSize;
    U32                     HardwareAddressLow;
    U32                     HardwareAddressHigh;
    U32                     MaxWireSpeedLow;
    U32                     MaxWireSpeedHigh;
    U32                     BucketsRemaining;
    U32                     MaxReplySize;
} CONFIG_PAGE_LAN_1, *PTR_CONFIG_PAGE_LAN_1,
  LANPage1_t, *pLANPage1_t;

#define MPI_LAN_PAGE_1_PAGEVERSION                  (0x02)

#define MPI_LAN_PAGE_1_DEV_STATE_RESET              (0x00)
#define MPI_LAN_PAGE_1_DEV_STATE_OPERATIONAL        (0x01)

#endif

