/*
 *  linux/drivers/fusion/mpictl.c
 *      High performance SCSI / Fibre Channel device driver.
 *      For use with PCI adapter(s):
 *          LSI SYM409xx (Fibre Channel)
 *      running LSI Logic Fusion MPT (Message Passing Technology) firmware.
 *
 *      Fusion MPT Misc Device Driver
 *
 *  Credits:
 *      This driver would not exist if not for Alan Cox's development
 *      of the linux i2o driver.
 *      A huge debt of gratitude is owed to David S. Miller (DaveM)
 *      for fixing much of the stupid and broken stuff in the early
 *      driver while porting to sparc64 platform.  THANK YOU!
 *      A big THANKS to Eddie C. Dost for fixing the ioctl path
 *      and most importantly f/w download on sparc64 platform!
 *      (plus Eddie's other helpful hints and insights)
 *
 *  Copyright (c) 1999-2000 LSI Logic Corporation
 *  Originally By: Steven J. Ralston, Noah Romer
 *  (mailto:Steve.Ralston@lsil.com)
 *
 *  $Id: mpictl.c,v 1.49 2000/09/27 10:17:01 sralston Exp $
 */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    NO WARRANTY
    THE PROGRAM IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OR
    CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED INCLUDING, WITHOUT
    LIMITATION, ANY WARRANTIES OR CONDITIONS OF TITLE, NON-INFRINGEMENT,
    MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. Each Recipient is
    solely responsible for determining the appropriateness of using and
    distributing the Program and assumes all risks associated with its
    exercise of rights under this Agreement, including but not limited to
    the risks and costs of program errors, damage to or loss of data,
    programs or equipment, and unavailability or interruption of operations.

    DISCLAIMER OF LIABILITY
    NEITHER RECIPIENT NOR ANY CONTRIBUTORS SHALL HAVE ANY LIABILITY FOR ANY
    DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING WITHOUT LIMITATION LOST PROFITS), HOWEVER CAUSED AND
    ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
    USE OR DISTRIBUTION OF THE PROGRAM OR THE EXERCISE OF ANY RIGHTS GRANTED
    HEREUNDER, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

// GRRRRRRRRRR...  FIXME!!!
#if defined(MODULE) && !defined(__GENKSYMS__)
#include <linux/module.h>
#endif

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/malloc.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
#include <linux/spinlock.h>
#endif

#include <asm/uaccess.h>
#include <asm/io.h>

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#define CoPyRiGhT "Copyright (c) 1999-2000 LSI Logic Corporation"
#define MoDuLeAuThOr  "Steven J. Ralston, Noah Romer"
#include "mpi.h"

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
#define my_NAME		"Fusion MPT misc device driver"
#define my_VERSION	LINUX_MPI_VERSION_EXP

#ifdef MODULE

EXPORT_NO_SYMBOLS;
MODULE_AUTHOR(MoDuLeAuThOr);
MODULE_DESCRIPTION(my_NAME);

#endif

//  Just in case...
#ifndef MPI_MINOR
#define MPI_MINOR  167
#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

static int mpictl_id = -1;
static int rwperf_reset = 0;

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

static int mpi_ioctl_rwperf(unsigned long arg);
static int mpi_ioctl_rwperf_status(unsigned long arg);
static int mpi_ioctl_rwperf_reset(unsigned long arg);
static int mpi_ioctl_fw_download(unsigned long arg);
static int mpictl_do_fw_download(int ioc, char *ufwbuf, size_t fwlen);
static int mpi_ioctl_scsi_cmd(unsigned long arg);

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#define myMAX_TARGETS	(1<<4)
#define myMAX_LUNS	(1<<3)
#define myMAX_T_MASK	(myMAX_TARGETS-1)
#define myMAX_L_MASK	(myMAX_LUNS-1)
static U8  DevInUse[myMAX_TARGETS][myMAX_LUNS] = {{0,0}};
static U32 DevIosCount[myMAX_TARGETS][myMAX_LUNS] = {{0,0}};

static U32 fwReplyBuffer[16];
static pMPIDefaultReply_t ReplyMsg = NULL;

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This is the callback for any message we have posted. The message itself
 *  will be returned to the message pool when we return from the IRQ
 *
 *  This runs in irq context so be short and sweet.
 */
static int
mpictl_reply(MPI_ADAPTER *ioc, MPI_FRAME_HDR *req, MPI_FRAME_HDR *reply)
{
	U8 targ;

	//dprintk((KERN_DEBUG "mpictl: Got mpictl_reply()!\n"));

	if (req && req->u.hdr.Function == MPI_FUNCTION_SCSI_IO_REQUEST) {
		targ = req->u.scsireq.TargetID & myMAX_T_MASK;
		DevIosCount[targ][0]--;
	} else if (reply && req && req->u.hdr.Function == MPI_FUNCTION_FW_DOWNLOAD) {
		// NOTE: Expects/requires non-Turbo reply!
		dprintk((KERN_INFO "mpictl: Caching MPI_FUNCTION_FW_DOWNLOAD reply!\n"));
		memcpy(fwReplyBuffer, reply, MIN(sizeof(fwReplyBuffer), 4*reply->u.reply.MsgLength));
		ReplyMsg = (pMPIDefaultReply_t) fwReplyBuffer;
	}

	return 1;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static long long
ctl_llseek(struct file *file, long long offset, int origin)
{
	return -ESPIPE;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static ssize_t
ctl_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	printk(KERN_ERR "mpictl: mpictl WRITE not yet supported\n");
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static ssize_t
ctl_read(struct file *file, char *buf, size_t count, loff_t *ptr)
{
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/* All of the mpictl commands can potentially sleep, which is illegal
 * with a spinlock held, thus we perform mutual exclusion in the following
 * way.
 */
static DECLARE_WAIT_QUEUE_HEAD(mpictl_wait);
static spinlock_t mpictl_mutex = SPIN_LOCK_UNLOCKED;
static int mpictl_users = 0;

static void mpictl_lock(void)
{
	DECLARE_WAITQUEUE(wait, current);

	add_wait_queue(&mpictl_wait, &wait);
	current->state = TASK_UNINTERRUPTIBLE;
	while (1) {
		spin_lock(&mpictl_mutex);
		if (!mpictl_users) {
			mpictl_users = 1;
			spin_unlock(&mpictl_mutex);
			break;
		}
		spin_unlock(&mpictl_mutex);
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&mpictl_wait, &wait);
}

static void mpictl_unlock(void)
{
	spin_lock(&mpictl_mutex);
	mpictl_users = 0;
	wake_up(&mpictl_wait);
	spin_unlock(&mpictl_mutex);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  MPI ioctl handler
 */
static int
ctl_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret;

	mpictl_lock();

	switch(cmd)
	{
		case MPIRWPERF:
			ret = mpi_ioctl_rwperf(arg);
			break;

		case MPIRWPERF_CHK:
			ret = mpi_ioctl_rwperf_status(arg);
			break;

		case MPIRWPERF_RESET:
			ret = mpi_ioctl_rwperf_reset(arg);
			break;

		case MPIFWDOWNLOAD:
			ret = mpi_ioctl_fw_download(arg);
			break;

		case MPISCSICMD:
			ret = mpi_ioctl_scsi_cmd(arg);
			break;

		default:
			ret = -EINVAL;
	}

	mpictl_unlock();

	return ret;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
ctl_open(struct inode *inode, struct file *file)
{
	/*
	 * Should support multiple management users
	 */
	MOD_INC_USE_COUNT;
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
ctl_release(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
mpi_ioctl_fw_download(unsigned long arg)
{
	struct mpi_fw_xfer	*ufwdl = (struct mpi_fw_xfer *) arg;
	struct mpi_fw_xfer	 kfwdl;

	dprintk((KERN_INFO "mpi_ioctl_fwdl called. mpictl_id = %xh\n", mpictl_id)); //tc
	if (copy_from_user(&kfwdl, ufwdl, sizeof(struct mpi_fw_xfer)) < 0) {
		printk(KERN_ERR "%s@%d::mpi_ioctl_fwdl - "
				"Unable to copy mpi_fw_xfer struct @ %p\n",
				__FILE__, __LINE__, (void*)ufwdl);
		return -EFAULT;
	}

	return mpictl_do_fw_download(kfwdl.ioc, kfwdl.bufp, kfwdl.fwlen);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 * MPI FW Download
 */
static int
mpictl_do_fw_download(int ioc, char *ufwbuf, size_t fwlen)
{
	FWDownload_t		*dlmsg;
	MPI_FRAME_HDR		*mf;
	MPI_ADAPTER		*iocp;
	char			*fwbuf;
	dma_addr_t		 fwbuf_dma;
	FWDownloadTCSGE_t	*fwVoodoo;
	SGEAllUnion_t		*fwSgl;

	dprintk((KERN_INFO "mpictl_do_fwdl called. mpictl_id = %xh.\n", mpictl_id)); //tc

//dprintk((KERN_INFO "DbG: kfwdl.bufp  = %p\n", ufwbuf));
//dprintk((KERN_INFO "DbG: kfwdl.fwlen = %d\n", (int)fwlen));
//dprintk((KERN_INFO "DbG: kfwdl.ioc   = %04xh\n", ioc));
	/*
	 *  Need to kmalloc fwbuf
	 * FIXME! This breaks as soon as IOC f/w goes above 128 Kb!
	 *  (it's at 113-123 Kb right now!!!)
	 */
	if ((fwbuf = kmalloc(fwlen, GFP_USER)) == NULL) {
		printk(KERN_ERR "%s@%d::mpi_ioctl_fwdl - "
				"Unable to alloc %d byte f/w buffer\n",
				__FILE__, __LINE__, (int)fwlen);
		return -ENOMEM;
	}
//dprintk((KERN_INFO "DbG: fwbuf       = %p\n", fwbuf));

	if (copy_from_user(fwbuf, ufwbuf, fwlen) < 0) {
		printk(KERN_ERR "%s@%d::mpi_ioctl_fwdl - "
				"Unable to copy f/w buffer @ %p\n",
				__FILE__, __LINE__, (void*)ufwbuf);
		return -EFAULT;
	}

	/* Perform Firmware download.
	 */

	if ((ioc = mpi_verify_adapter(ioc, &iocp)) < 0) {
		printk("%s@%d::mpi_ioctl_fwdl - ioc%d not found!\n",
				__FILE__, __LINE__, ioc);
		return -ENXIO;			  /* (-6) No such device or address */
	}

	mf = MpiGetMsgFrame(mpictl_id, ioc);

	dlmsg = (FWDownload_t*) mf;
	dlmsg->Reserved[0] = 0;
	dlmsg->Reserved[1] = 0;
	dlmsg->ChainOffset = 0;
	dlmsg->Function = MPI_FUNCTION_FW_DOWNLOAD;
	dlmsg->Reserved1[0] = dlmsg->Reserved1[1] = dlmsg->Reserved1[2] = 0;
	dlmsg->MsgFlags = 0;

	fwVoodoo = (FWDownloadTCSGE_t *) &dlmsg->SGL;
	fwVoodoo->Reserved = 0;
	fwVoodoo->ContextSize = 0;
	fwVoodoo->DetailsLength = 12;
	fwVoodoo->Flags = MPI_SGE_FLAGS_TRANSACTION_ELEMENT;
	fwVoodoo->Checksum = 0;
	fwVoodoo->ImageOffset = 0;
	fwVoodoo->ImageSize = cpu_to_le32(fwlen);

	fwSgl = (SGEAllUnion_t *) (fwVoodoo + 1);
	fwSgl->u.Simple.FlagsLength = cpu_to_le32(
			((MPI_SGE_FLAGS_LAST_ELEMENT |
			  MPI_SGE_FLAGS_END_OF_BUFFER |
			  MPI_SGE_FLAGS_END_OF_LIST |
			  MPI_SGE_FLAGS_SIMPLE_ELEMENT |
			  MPI_SGE_FLAGS_SYSTEM_ADDRESS |
			  MPI_SGE_FLAGS_32_BIT_ADDRESSING |
			  MPI_SGE_FLAGS_32_BIT_CONTEXT) << MPI_SGE_FLAGS_SHIFT) |
			 fwlen );
	fwbuf_dma = pci_map_single(iocp->pcidev, fwbuf, fwlen, PCI_DMA_TODEVICE);
	fwSgl->u.Simple.u.Address32 = cpu_to_le32(fwbuf_dma);


	ReplyMsg = NULL;
	MpiPutMsgFrame(mpictl_id, ioc, mf);

	/*
	 *  Wait until the reply has been received
	 */
	{
		int	 foo = 0;

		while (ReplyMsg == NULL) {
			if (!(foo%1000000)) {
				dprintk((KERN_INFO "DbG::mpictl_do_fwdl: "
					   "In ReplyMsg loop - iteration %d\n",
					   foo)); //tc
			}
			if (++foo > 60000000)
				return -ETIME;
			mb();
			schedule();
			barrier();
		}
	}
	
	pci_unmap_single(iocp->pcidev, fwbuf_dma, fwlen, PCI_DMA_TODEVICE);
	if (fwbuf)
		kfree(fwbuf);

	if (le16_to_cpu(ReplyMsg->IOCStatus) == MPI_IOCSTATUS_SUCCESS) {
		printk(KERN_INFO "mpi_ioctl_fwdl: F/W successfully sent to HAB!\n");
		return 0;
	} else if (le16_to_cpu(ReplyMsg->IOCStatus) == MPI_IOCSTATUS_INVALID_FUNCTION) {
		printk(KERN_WARNING "mpi_ioctl_fwdl: Uh-Oh! The HAB didn't recognize the function number.\n");
		printk(KERN_WARNING "Time to go bang on somebodies door.\n");
		return -EBADRQC;
	} else if (le16_to_cpu(ReplyMsg->IOCStatus) == MPI_IOCSTATUS_BUSY) {
		printk(KERN_WARNING "mpi_ioctl_fwdl: IOC busy. Try again later.\n");
		return -EBUSY;
	} else {
		printk(KERN_WARNING "mpi_ioctl_fwdl: IOC returned status of %d. Bad Voodoo.\n",
				    le16_to_cpu(ReplyMsg->IOCStatus));
		return -ENOMSG;
	}
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  NEW rwperf (read/write performance) stuff starts here...
 */

//#define MAX_SCSI_FRAGS	9
#define MAX_FRAGS_SPILL1	9
#define MAX_FRAGS_SPILL2	15
#define FRAGS_PER_BUCKET	(MAX_FRAGS_SPILL2 + 1)

//#define MAX_CHAIN_FRAGS	64
//#define MAX_CHAIN_FRAGS	(15+15+15+16)
#define MAX_CHAIN_FRAGS		(4 * MAX_FRAGS_SPILL2 + 1)

//  Define max sg LIST bytes ( == (#frags + #chains) * 8 bytes each)
//  Works out to: 592d bytes!     (9+1)*8 + 4*(15+1)*8
//                  ^----------------- 80 + 512
#define MAX_SGL_BYTES		((MAX_FRAGS_SPILL1 + 1 + (4 * FRAGS_PER_BUCKET)) * 8)

/* linux will only ever give a max of 128Kb contiguous mem bytes */
#define MAX_KMALLOC_SZ		(128*1024)

struct buflist {
	U8	*kptr;
	int	 len;
};

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static SGESimple32_t *
kbuf_alloc_2_sgl(int bytes, U32 dir, int *frags,
		 struct buflist **blp, dma_addr_t *sglbuf_dma, MPI_ADAPTER *ioc)
{
	SGESimple32_t	*sglbuf = NULL;
	struct buflist	*buflist = NULL;
	int		 numfrags = 0;
	int		 fragcnt = 0;
	int		 alloc_sz = MIN(bytes,MAX_KMALLOC_SZ);	// avoid kernel warning msg!
	int		 bytes_allocd = 0;
	int		 this_alloc;
	SGESimple32_t	*sgl;
	U32		 pa;					// phys addr
	SGEChain32_t	*last_chain = NULL;
	SGEChain32_t	*old_chain = NULL;
	int		 chaincnt = 0;
	int		 i, buflist_ent;
	int		 sg_spill = MAX_FRAGS_SPILL1;

	*frags = 0;
	*blp = NULL;
	i = MAX_SGL_BYTES / 8;
	buflist = kmalloc(i, GFP_USER);
	if (buflist == NULL)
		return NULL;
	memset(buflist, 0, i);
	buflist_ent = 0;

	sglbuf = pci_alloc_consistent(ioc->pcidev, MAX_SGL_BYTES, sglbuf_dma);
	if (sglbuf == NULL)
		goto free_and_fail;

	sgl = sglbuf;
	while (bytes_allocd < bytes) {
		this_alloc = MIN(alloc_sz, bytes-bytes_allocd);
		buflist[buflist_ent].len = this_alloc;
		buflist[buflist_ent].kptr = pci_alloc_consistent(ioc->pcidev,
								 this_alloc,
								 &pa);
		if (buflist[buflist_ent].kptr == NULL) {
			alloc_sz = alloc_sz / 2;
			if (alloc_sz == 0) {
				printk(KERN_WARNING "mpictl-SG: No can do - "
						    "not enough memory!   :-(\n");
				printk(KERN_WARNING "mpictl-SG: (freeing %d frags)\n",
						    numfrags);
				goto free_and_fail;
			}
			continue;
		} else {
			bytes_allocd += this_alloc;

			/* Write one SIMPLE sge */
			sgl->FlagsLength = cpu_to_le32(0x10000000|dir|this_alloc);
			sgl->Address = cpu_to_le32(pa);
			fragcnt++;
			numfrags++;
			sgl++;
			buflist_ent++;
		}

		if (bytes_allocd >= bytes)
			break;

		/* Need to chain? */
		if (fragcnt == sg_spill) {
			dma_addr_t chain_link;

			if (last_chain != NULL)
				last_chain->NextChainOffset = 0x1E;

			fragcnt = 0;
			sg_spill = MAX_FRAGS_SPILL2;

			/* fixup previous SIMPLE sge */
			sgl[-1].FlagsLength |= cpu_to_le32(0x80000000);

			chain_link = (*sglbuf_dma) +
				((U8 *)(sgl+1) - (U8 *)sglbuf);

			/* Write one CHAIN sge */
			sgl->FlagsLength = cpu_to_le32(0x30000080);
			sgl->Address = cpu_to_le32(chain_link);

			old_chain = last_chain;
			last_chain = (SGEChain32_t*)sgl;
			chaincnt++;
			numfrags++;
			sgl++;
		}

		/* overflow check... */
		if (numfrags*8 > MAX_SGL_BYTES) {
			/* GRRRRR... */
			printk(KERN_WARNING "mpictl-SG: No can do - "
					    "too many SG frags!   :-(\n");
			printk(KERN_WARNING "mpictl-SG: (freeing %d frags)\n",
					    numfrags);
			goto free_and_fail;
		}
	}

	/* Last sge fixup: set LE+eol+eob bits */
	sgl[-1].FlagsLength |= cpu_to_le32(0xC1000000);

	/* Chain fixup needed? */
	if (last_chain != NULL && fragcnt < 16)
		last_chain->Length = cpu_to_le16(fragcnt * 8);

	*frags = numfrags;
	*blp = buflist;

	dprintk((KERN_INFO "mpictl-SG: kbuf_alloc_2_sgl() - "
			   "%d SG frags generated!  (%d CHAIN%s)\n",
			   numfrags, chaincnt, chaincnt>1?"s":""));

	dprintk((KERN_INFO "mpictl-SG: kbuf_alloc_2_sgl() - "
			   "last (big) alloc_sz=%d\n",
			   alloc_sz));

	return sglbuf;

free_and_fail:
	if (sglbuf != NULL) {
		int i;

		for (i = 0; i < numfrags; i++) {
			dma_addr_t dma_addr;
			U8 *kptr;
			int len;

			if ((sglbuf[i].FlagsLength >> 24) == 0x30)
				continue;

			dma_addr = le32_to_cpu(sglbuf[i].Address);
			kptr = buflist[i].kptr;
			len = buflist[i].len;

			pci_free_consistent(ioc->pcidev, len, kptr, dma_addr);
		}
		pci_free_consistent(ioc->pcidev, MAX_SGL_BYTES, sglbuf, *sglbuf_dma);
	}
	kfree(buflist);
	return NULL;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static void
kfree_sgl(SGESimple32_t *sgl, dma_addr_t sgl_dma, struct buflist *buflist, MPI_ADAPTER *ioc)
{
	SGESimple32_t	*sg = sgl;
	struct buflist	*bl = buflist;
	U32		 nib;
	int		 n = 0;

	nib = (sg->FlagsLength & 0xF0000000) >> 28;
	while (! (nib & 0x4)) { /* eob */
		/* skip ignore/chain. */
		if (nib == 0 || nib == 3) {
			;
		} else if (sg->Address) {
			dma_addr_t dma_addr;
			void *kptr;
			int len;

			dma_addr = le32_to_cpu(sg->Address);
			kptr = bl->kptr;
			len = bl->len;
			pci_free_consistent(ioc->pcidev, len, kptr, dma_addr);
			n++;
		}
		sg++;
		bl++;
		nib = (sg->FlagsLength & 0xF0000000) >> 28;
	}

	/* we're at eob! */
	if (sg->Address) {
		dma_addr_t dma_addr;
		void *kptr;
		int len;

		dma_addr = le32_to_cpu(sg->Address);
		kptr = bl->kptr;
		len = bl->len;
		pci_free_consistent(ioc->pcidev, len, kptr, dma_addr);
		n++;
	}

	pci_free_consistent(ioc->pcidev, MAX_SGL_BYTES, sgl, sgl_dma);
	kfree(buflist);
	dprintk((KERN_INFO "mpictl-SG: Free'd 1 SGL buf + %d kbufs!\n", n));
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
mpi_ioctl_rwperf_init(struct mpi_raw_r_w *dest, unsigned long src,
		      char *caller, MPI_ADAPTER **iocpp)
{
	char	*myname = "mpi_ioctl_rwperf_init()";
	int	 ioc;


	/* get copy of structure passed from user space */
	if (copy_from_user(dest, (void*)src, sizeof(*dest))) {
		printk(KERN_ERR "%s@%d::%s - Can't copy mpi_raw_r_w data @ %p\n",
				__FILE__, __LINE__, myname, (void*)src);
		return -EFAULT;				/* (-14) Bad address */
	} else {
		dprintk((KERN_INFO "mpictl-perf: PerfInfo.{ioc,b:t:l,qd,iters,nblks}"
				   ": %02xh %06xh %d %d %d\n",
				   dest->iocport, dest->btl,
				   (int)dest->qdepth, dest->iters, dest->nblks ));
		dprintk((KERN_INFO "mpictl-perf: PerfInfo.{cache,skip,range,rdwr,seqran}"
				   ": %d %d %d %d %d\n",
				   dest->cache_sz, dest->skip, dest->range,
				   dest->rdwr, dest->seqran ));

		/* Get the MPI adapter id. */
		if ((ioc = mpi_verify_adapter((dest->iocport & 0x0000FF00) >> 8, iocpp)) < 0) {
			printk(KERN_ERR "%s@%d::%s - ioc%2x/%06x not found!\n",
					__FILE__, __LINE__, myname,
					dest->iocport, dest->btl );
			return -ENXIO;			/* (-6) No such device or address */
		} else {
			dprintk(("mpictl-perf: %s using mpi/ioc%x/%06x\n",
					caller, dest->iocport, dest->btl ));
		}
	}

	return ioc;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*  Treat first N blocks of disk as sacred!  */
#define SACRED_BLOCKS	100

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
mpi_ioctl_rwperf(unsigned long arg)
{
	struct mpi_raw_r_w	 kPerfInfo;
				/* NOTE: local copy, on stack==KERNEL_SPACE! */
	U8		 bus;
	U8		 target, targetM;
	U8		 lun, lunM;
	U8		 scsiop;
	int		 qdepth;
	int		 iters;
	int		 cache_sz;
	U32		 xferbytes;
	U32		 scsidir;
	U32		 qtag;
	U32		 scsictl;
	U32		 sgdir;
	U32		 blkno;
	U32		 sbphys;
	SGESimple32_t	*sgl;
	dma_addr_t	 sgl_dma;
	struct buflist	*buflist;
	SGESimple32_t	*sgOut, *sgIn;
	int		 numfrags;
	U32		*msg;
	int		 i;
	int		 ioc;
	MPI_FRAME_HDR	*mf;
	MPI_ADAPTER	*iocp;
	int		 sgfragcpycnt;
	int		 blklo, blkhi;
	U8		 nextchainoffset;
	U8		*SenseBuf;
	dma_addr_t	 SenseBufDMA;
	char		*myname = "ctl_rwperf()";

    dprintk((KERN_INFO "%s - starting...\n", myname));

    /* Validate target device */
    if ((ioc = mpi_ioctl_rwperf_init(&kPerfInfo, arg, myname, &iocp)) < 0)
        return ioc;

    /* Allocate DMA'able memory for the sense buffer. */
    SenseBuf = pci_alloc_consistent(iocp->pcidev, 256, &SenseBufDMA);

    /* set perf parameters from input */
    bus = (kPerfInfo.iocport & 0x000000FF);
    target = (kPerfInfo.btl & 0x0000FF00) >> 8;
    targetM = target & myMAX_T_MASK;
    lun = (kPerfInfo.btl & 0x000000FF);
    lunM = lun & myMAX_L_MASK;
    qdepth = kPerfInfo.qdepth;
    iters = kPerfInfo.iters;
    xferbytes = ((U32)kPerfInfo.nblks)<<9;

    DevInUse[targetM][lunM] = 1;
    DevIosCount[targetM][lunM] = 0;

    cache_sz = kPerfInfo.cache_sz * 1024;	// CacheSz in Kb!

    /* ToDo: */
    /* get capacity (?) */


    // pre-build, one time, everything we can for speed in the loops below...

    scsiop = 0x28;				// default to READ!
    scsidir = MPI_SCSIIO_CONTROL_READ;		// DATA IN  (host<--ioc<--dev)
						// 02000000
    qtag = MPI_SCSIIO_CONTROL_SIMPLEQ;		// 00000000

    if (xferbytes == 0) {
        // Do 0-byte READ!!!
        //  IMPORTANT!  Need to set no SCSI DIR for this!
        scsidir = MPI_SCSIIO_CONTROL_NODATATRANSFER;
    }

    scsictl = scsidir | qtag;

    /*
     *  Set sgdir for DMA transfer.
     */
//    sgdir   = 0x04000000;		// WRITE
    sgdir = 0x00000000;			// READ

    if ((sgl = kbuf_alloc_2_sgl(MAX(512,xferbytes), sgdir, &numfrags, &buflist, &sgl_dma, iocp)) == NULL)
        return -ENOMEM;

    sgfragcpycnt = MIN(10,numfrags);
    nextchainoffset = 0;
    if (numfrags > 10)
        nextchainoffset = 0x1E;

    sbphys = SenseBufDMA;

    rwperf_reset = 0;

//    do {	// target-loop

        blkno = SACRED_BLOCKS;		// Treat first N blocks as sacred!
					// FIXME!  Skip option
        blklo = blkno;
        blkhi = blkno;

        do {    // inner-loop

            while ((mf = MpiGetMsgFrame(mpictl_id, ioc)) == NULL) {
                mb();
                schedule();
                barrier();
            }
            msg = (U32*)mf;

            /* Start piecing the SCSIIORequest together */
            msg[0] = 0x00000000 | nextchainoffset<<16 | target;
            msg[1] = 0x0000FF0A;				// 255 sense bytes, 10-byte CDB!
            msg[3] = lun << 8;
            msg[4] = 0;
            msg[5] = scsictl;

            // 16 bytes of CDB @ msg[6,7,8,9] are below...

            msg[6] = (   ((blkno & 0xFF000000) >> 8)
                       | ((blkno & 0x00FF0000) << 8)
                       | scsiop );
            msg[7] = (   (((U32)kPerfInfo.nblks & 0x0000FF00) << 16)
                       | ((blkno & 0x000000FF) << 8)
                       | ((blkno & 0x0000FF00) >> 8) );
            msg[8] = (kPerfInfo.nblks & 0x00FF);
            msg[9] = 0;

            msg[10] = xferbytes;

//            msg[11] = 0xD0000100;
//            msg[12] = sbphys;
//            msg[13] = 0;
            msg[11] = sbphys;

            // Copy the SGL...
            if (xferbytes) {
                sgOut = (SGESimple32_t*)&msg[12];
                sgIn  = sgl;
                for (i=0; i < sgfragcpycnt; i++)
                    *sgOut++ = *sgIn++;
            }

            // fubar!  QueueDepth issue!!!
            while (    !rwperf_reset
                    && (DevIosCount[targetM][lunM] >= MIN(qdepth,64)) )
            {
                mb();
                schedule();
                barrier();
            }

//            blkno += kPerfInfo.nblks;
// EXP Stuff!
// Try optimizing to certain cache size for the target!
// by keeping blkno within cache range if at all possible
#if 0
            if (    cache_sz
                 && ((2 * kPerfInfo.nblks) <= (cache_sz>>9))
                 && ((blkno + kPerfInfo.nblks) > ((cache_sz>>9) + SACRED_BLOCKS)) )
                blkno = SACRED_BLOCKS;
            else
                blkno += kPerfInfo.nblks;
#endif
// Ok, cheat!
            if (cache_sz && ((blkno + kPerfInfo.nblks) > ((cache_sz>>9) + SACRED_BLOCKS)) )
                   blkno = SACRED_BLOCKS;
            else
                blkno += kPerfInfo.nblks;

            if (blkno > blkhi)
                blkhi = blkno;

            DevIosCount[targetM][lunM]++;

            /*
             *  Finally, post the request
             */
            MpiPutMsgFrame(mpictl_id, ioc, mf);


            /* let linux breath! */
            mb();
            schedule();
            barrier();

            //dprintk((KERN_DEBUG "mpictl-perf: inner-loop, cnt=%d\n", iters));

        } while ((--iters > 0) && !rwperf_reset);

        dprintk((KERN_INFO "mpictl-perf: DbG: blklo=%d, blkhi=%d\n", blklo, blkhi));
        dprintk((KERN_INFO "mpictl-perf: target-loop, thisTarget=%d\n", target));

//        //  TEMPORARY!
//        target = 0;

//    } while (target);


    if (DevIosCount[targetM][lunM]) {
        dprintk((KERN_INFO "  DbG: DevIosCount[%d][%d]=%d\n",
                targetM, lunM, DevIosCount[targetM][lunM]));
    }

    while (DevIosCount[targetM][lunM]) {
        //dprintk((KERN_DEBUG "  DbG: Waiting... DevIosCount[%d][%d]=%d\n",
        //        targetM, lunM, DevIosCount[targetM][lunM]));
        mb();
        schedule();
        barrier();
    }
    DevInUse[targetM][lunM] = 0;

    pci_free_consistent(iocp->pcidev, 256, SenseBuf, SenseBufDMA);

    if (sgl)
        kfree_sgl(sgl, sgl_dma, buflist, iocp);

    dprintk((KERN_INFO "  *** done ***\n"));

    return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
mpi_ioctl_rwperf_status(unsigned long arg)
{
	struct mpi_raw_r_w	 kPerfInfo;
				/* NOTE: local copy, on stack==KERNEL_SPACE! */
	MPI_ADAPTER	*iocp;
	int		 ioc;
//	U8		 targ;
//	U8		 lun;
	int		 T, L;
	char		*myname = "mpi_ioctl_rwperf_status()";


	dprintk((KERN_INFO "%s - starting...\n", myname));

	/* Get a pointer to the MPI adapter. */
	if ((ioc = mpi_ioctl_rwperf_init(&kPerfInfo, arg, myname, &iocp)) < 0)
		return ioc;

	/* set perf parameters from input */
//	targ = (kPerfInfo.btl & 0x0000FF00) >> 8;
//	lun = (kPerfInfo.btl & 0x000000FF);

	for (T=0; T < myMAX_TARGETS; T++)
		for (L=0; L < myMAX_LUNS; L++)
			if (DevIosCount[T][L]) {
				printk(KERN_INFO "%s: ioc%d->00:%02x:%02x"
						 ", IosCnt=%d\n",
						 myname, ioc, T, L, DevIosCount[T][L] );
			}

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
mpi_ioctl_rwperf_reset(unsigned long arg)
{
	struct mpi_raw_r_w	 kPerfInfo;
				/* NOTE: local copy, on stack==KERNEL_SPACE! */
	MPI_ADAPTER	*iocp;
	int		 ioc;
//	U8		 targ;
//	U8		 lun;
	int		 T, L;
	int		 i;
	char		*myname = "mpi_ioctl_rwperf_reset()";

	dprintk((KERN_INFO "%s - starting...\n", myname));

	/* Get MPI adapter id. */
	if ((ioc = mpi_ioctl_rwperf_init(&kPerfInfo, arg, myname, &iocp)) < 0)
		return ioc;

	/* set perf parameters from input */
//	targ = (kPerfInfo.btl & 0x0000FF00) >> 8;
//	lun = (kPerfInfo.btl & 0x000000FF);

	rwperf_reset = 1;
	for (i=0; i < 1000000; i++) {
		mb();
		schedule();
		barrier();
	}
	rwperf_reset = 0;

	for (T=0; T < myMAX_TARGETS; T++)
		for (L=0; L < myMAX_LUNS; L++)
			if (DevIosCount[T][L]) {
				printk(KERN_INFO "%s: ioc%d->00:%02x:%02x, "
						 "IosCnt RESET! (from %d to 0)\n",
						 myname, ioc, T, L, DevIosCount[T][L] );
				DevIosCount[T][L] = 0;
				DevInUse[T][L] = 0;
			}

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
mpi_ioctl_scsi_cmd(unsigned long arg)
{
	return -ENOSYS;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

static struct file_operations mpictl_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	owner:		THIS_MODULE,
#endif
	llseek:		ctl_llseek,
	read:		ctl_read,
	write:		ctl_write,
	ioctl:		ctl_ioctl,
	open:		ctl_open,
	release:	ctl_release,
};

static struct miscdevice mpictl_miscdev = {
	MPI_MINOR,
	"mpictl",
	&mpictl_fops
};

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#if defined(__sparc__) && defined(__sparc_v9__)		/*{*/

/* The dynamic ioctl32 compat. registry only exists in >2.3.x sparc64 kernels */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)		/*{*/
extern int register_ioctl32_conversion(unsigned int cmd,
				       int (*handler)(unsigned int,
						      unsigned int,
						      unsigned long,
						      struct file *));
int unregister_ioctl32_conversion(unsigned int cmd);

struct mpi_fw_xfer32 {
	u32 bufp;
	unsigned int fwlen;
	int ioc;
};

#define MPIFWDOWNLOAD32     _IOWR(MPI_MAGIC_NUMBER,15,struct mpi_fw_xfer32)

extern asmlinkage int sys_ioctl(unsigned int fd, unsigned int cmd, unsigned long arg);

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
sparc32_mpifwxfer_ioctl(unsigned int fd, unsigned int cmd,
                        unsigned long arg, struct file *filp)
{
	struct mpi_fw_xfer32 fw32;
	struct mpi_fw_xfer kfw;

	if (copy_from_user(&fw32, (char *)arg, sizeof(fw32)))
		return -EFAULT;

	kfw.bufp = (void *)(unsigned long)fw32.bufp;
	kfw.fwlen = fw32.fwlen;
	kfw.ioc = fw32.ioc;

	return mpictl_do_fw_download(kfw.ioc, kfw.bufp, kfw.fwlen);
}

#if 0		//{
static int sparc32_mpifwxfer_ioctl(unsigned int fd,
				   unsigned int cmd,
				   unsigned long arg,
				   struct file *filp)
{
	struct mpi_fw_xfer32 fw32;
	struct mpi_fw_xfer kfw;
	mm_segment_t old_fs;
	int ret;

	if (copy_from_user(&fw32, (char *)arg, sizeof(fw32)))
		return -EFAULT;

	kfw.bufp = (void *)(unsigned long)fw32.bufp;
	kfw.fwlen = fw32.fwlen;
	kfw.ioc = fw32.ioc;

	old_fs = get_fs();
	set_fs(KERNEL_DS);
	ret = sys_ioctl(fd, MPIFWDOWNLOAD, (unsigned long)&kfw);
	set_fs(old_fs);

	return ret;
}
#endif		/*} 0 */

#endif		/*} linux >= 2.3.x */
#endif		/*} sparc */

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
#ifdef MODULE
int init_module(void)
#else
int __init mpictl_init(void)
#endif
{
	int err;
	int where = 1;

	show_mpimod_ver(my_NAME, my_VERSION);

#if defined(__sparc__) && defined(__sparc_v9__)		/*{*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)		/*{*/
	err = register_ioctl32_conversion(MPIRWPERF, NULL);
	if (++where && err) goto out_fail;
	err = register_ioctl32_conversion(MPIRWPERF_CHK, NULL);
	if (++where && err) goto out_fail;
	err = register_ioctl32_conversion(MPIRWPERF_RESET, NULL);
	if (++where && err) goto out_fail;
	err = register_ioctl32_conversion(MPIFWDOWNLOAD32,
					  sparc32_mpifwxfer_ioctl);
	if (++where && err) goto out_fail;
#endif		/*} linux >= 2.3.x */
#endif		/*} sparc */

	if (misc_register(&mpictl_miscdev) == -1) {
		printk(KERN_ERR "mpictl: can't register misc device [%d].\n", MPI_MINOR);
		err = -EBUSY;
		goto out_fail;
	}
	printk(KERN_INFO "Registered Fusion MPT misc device driver \"%s\" at (major,minor) %d,%d\n",
			 mpictl_miscdev.name, MISC_MAJOR, mpictl_miscdev.minor);

	/*
	 *  Install our handler
	 */
	++where;
	if ((mpictl_id = mpi_register(mpictl_reply)) < 0) {
		printk(KERN_ERR "mpictl: ERROR: Failed to register with Fusion MPT base driver\n");
		misc_deregister(&mpictl_miscdev);
		err = -EBUSY;
		goto out_fail;
	}

	return 0;

out_fail:

#if defined(__sparc__) && defined(__sparc_v9__)		/*{*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)		/*{*/
	printk(KERN_ERR "mpictl: ERROR: Failed to register ioctl32_conversion!"
			" (%d:err=%d)\n", where, err);
	unregister_ioctl32_conversion(MPIRWPERF);
	unregister_ioctl32_conversion(MPIRWPERF_CHK);
	unregister_ioctl32_conversion(MPIRWPERF_RESET);
	unregister_ioctl32_conversion(MPIFWDOWNLOAD32);
#endif		/*} linux >= 2.3.x */
#endif		/*} sparc */

	return err;
}

#ifdef MODULE

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
void cleanup_module(void)
{
	misc_deregister(&mpictl_miscdev);
	printk(KERN_INFO "Deregistered Fusion MPT misc device driver \"%s\" at (major,minor) %d,%d\n",
			mpictl_miscdev.name, MISC_MAJOR, mpictl_miscdev.minor);

	mpi_deregister(mpictl_id);
}

#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

