#include <stdio.h>
#include <unistd.h>
#include <string.h>
#define _GNU_SOURCE
#include <getopt.h>

#include "symsense.h"

#include "ascq_tbl.c"
#include "scsiops.c"
//#include "ascqops_tbl.c"

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#ifndef MIN
#define MIN(a,b)  ((a) <= (b) ? (a) : (b))
#endif

static u8 dummyInqData[SCSI_STD_INQUIRY_BYTES]
  = { 0x1F, 0x00, 0x00, 0x00,
      0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
static u8 dummySenseData[SCSI_STD_SENSE_BYTES]
  = { 0x70, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0A,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00 };
static u8 dummyCDB[16]
  = { 0xBE, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
static u8 dummyScsiData[16]
  = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

static ASCQ_Table_t *ASCQ_TablePtr = &ASCQ_Table[0];
#define ASCQ_TableSz  ASCQ_TableSize

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

static int   Generic_Sense(IO_Info_t *ioop);

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
int
usage(char *myname)
{
	fprintf(stderr, "%s: Usage:  %s [options]\n", myname, myname);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
int
main(int argc, char **argv)
{
	unsigned char	 sd[18];
	char		 scratch[1024];
	int 		 found;
	int 		 sdcnt;
	int		 len;
	int		 val;
	char		*s;
	char		*s2;
	int		 c;
	int		 digit_optind = 0;
	char		*myname;
	char		*infile = NULL;
	char		*outfile = NULL;
	int		 byteswap = 4;
	int		 periphtype = SCSI_TYPE_DAD;
	int		 sts = STS_CHECK_CONDITION;


	if ( (myname = strrchr(argv[0], '/')) == NULL ) {
		myname = argv[0];
	} else {
		myname++;
		argv[0] = myname;
	}

	/*  parse cmd-line args... */
	while (1) {
		int this_option_optind = optind ? optind : 1;
		int option_index = 0;
		static struct option long_options[] = {
			{"help", 0, NULL, 'h'},
			{"type", 1, NULL, 't'},
			{"scsistatus", 1, NULL, 's'},
			{"file", 1, NULL, 'f'},
			{"output", 1, NULL, 'o'},
			{"byteswap", 2, NULL, 'b'},
			{"debug", 2, NULL, 'd'},
			{"verbose", 0, NULL, 'v'},
			{NULL, 0, NULL, 0}
		};

		if ((c=getopt_long(argc, argv, "ht:s:f:o:bdv", long_options, &option_index)) == -1)
			break;

		switch (c) {
		case 'h': usage(myname); return 1;
		case 0:
			printf("option %s", long_options[option_index].name);
			if (optarg)
				printf(" with arg %s", optarg);
			printf("\n");
			break;
/*
		case '0':
		case '1':
		case '2':
			if (digit_optind != 0 && digit_optind != this_option_optind)
				printf("digits occur in two different argv-elements.\n");
			digit_optind = this_option_optind;
			printf("option %c\n", c);
			break;
*/
		case 't':
			if (sscanf(optarg, "%d", &periphtype) != 1)
				periphtype = SCSI_TYPE_DAD;
			printf("option t with arg `%s', value `%d'\n", optarg, periphtype);
			break;
		case 's':
			if (sscanf(optarg, "%d", &sts) != 1)
				sts = STS_CHECK_CONDITION;
			printf("option s with arg `%s', value `%d'\n", optarg, sts);
			break;
		case 'f':
			infile = optarg;
			printf("option f with value `%s'\n", optarg);
			break;
		case 'o':
			outfile = optarg;
			printf("option o with value `%s'\n", optarg);
			break;
		case 'b':
			printf("option b with value `%s'\n", optarg);
			break;
		case 'v':
			printf("option v\n");
			break;
		case 'd':
			printf("option d with value `%s'\n", optarg);
			break;
		case '?':
			break;
		default:
			printf ("?? getopt returned character code '%c' (%03o)??\n", c, c);
		}
	}

	if (optind < argc)
	{
		printf("non-option ARGV-elements:");
		while (optind < argc)
			printf(" %s", argv[optind++]);
		printf("\n");
	}

	dummyInqData[0] = periphtype;

	while ((s = fgets(scratch, sizeof(scratch), stdin)) != NULL) {
		int tmpsts;

		sdcnt = 0;
		tmpsts = 0x1FF;

		if ((s2 = strchr(s, (int)':')) != NULL) {
			(void) sscanf(s, "%x", &tmpsts);
			tmpsts &= 0x0FF;
			s = s2 + 1;
		}

		len = strlen(s);
//		fprintf(stderr, "DbG: len=%d, s=%s\n", len, s);
		s2 = s;

		do {
			found = sscanf(s, "%x", &val);
			if (found) {
//				fprintf(stderr, "DbG: %02X, found=%d\n", val, found);
				sd[sdcnt++] = val & 0x0FF;
				if ((s = strchr(s, (int)' ')) != NULL)
					s++;
			}
		} while (found && s && s < s2+len && sdcnt < 18);


		if (sdcnt >= 18) {
// EXP!
			char devFoo[32];
			IO_Info_t thisIo;
			int retry;

			thisIo.cdbPtr = dummyCDB;
			thisIo.sensePtr = sd;
			thisIo.SCSIStatus = (tmpsts==0x1FF) ? sts : tmpsts;
			thisIo.DoDisplay = 1;
			sprintf(devFoo, "ioc%d,scsi%d:%d", 0, 0, 0);
			thisIo.DevIDStr = devFoo;
			thisIo.inqPtr = dummyInqData;
			thisIo.dataPtr = dummyScsiData;

			retry = Generic_Sense(&thisIo);
		}
	}

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#define SPECIAL_ASCQ(c,q) \
	(((c) == 0x40 && (q) != 0x00) || ((c) == 0x4D) || ((c) == 0x70))

static char *ScsiStatusString[49] = {
        "GOOD",					/* 00h */
	"Bad-Reserved-01h",			/* 01h */
        "CHECK CONDITION",			/* 02h */
	"Bad-Reserved-03h",			/* 03h */
        "CONDITION MET",			/* 04h */
	"Bad-Reserved-05h",			/* 05h */
	"Bad-Reserved-06h",			/* 06h */
	"Bad-Reserved-07h",			/* 07h */
        "BUSY",					/* 08h */
	"Bad-Reserved-09h",			/* 09h */
	"Bad-Reserved-0Ah",			/* 0Ah */
	"Bad-Reserved-0Bh",			/* 0Bh */
	"Bad-Reserved-0Ch",			/* 0Ch */
	"Bad-Reserved-0Dh",			/* 0Dh */
	"Bad-Reserved-0Eh",			/* 0Eh */
	"Bad-Reserved-0Fh",			/* 0Fh */
        "INTERMEDIATE",				/* 10h */
	"Bad-Reserved-11h",			/* 11h */
	"Bad-Reserved-12h",			/* 12h */
	"Bad-Reserved-13h",			/* 13h */
        "INTERMEDIATE-CONDITION MET",		/* 14h */
	"Bad-Reserved-15h",			/* 15h */
	"Bad-Reserved-16h",			/* 16h */
	"Bad-Reserved-17h",			/* 17h */
        "RESERVATION CONFLICT",			/* 18h */
	"Bad-Reserved-19h",			/* 19h */
	"Bad-Reserved-1Ah",			/* 1Ah */
	"Bad-Reserved-1Bh",			/* 1Bh */
	"Bad-Reserved-1Ch",			/* 1Ch */
	"Bad-Reserved-1Dh",			/* 1Dh */
	"Bad-Reserved-1Eh",			/* 1Eh */
	"Bad-Reserved-1Fh",			/* 1Fh */
	"Bad-Reserved-20h",			/* 20h */
	"Bad-Reserved-21h",			/* 21h */
        "COMMAND TERMINATED",			/* 22h */
	"Bad-Reserved-23h",			/* 23h */
	"Bad-Reserved-24h",			/* 24h */
	"Bad-Reserved-25h",			/* 25h */
	"Bad-Reserved-26h",			/* 26h */
	"Bad-Reserved-27h",			/* 27h */
        "TASK SET FULL",			/* 28h */
	"Bad-Reserved-29h",			/* 29h */
	"Bad-Reserved-2Ah",			/* 2Ah */
	"Bad-Reserved-2Bh",			/* 2Bh */
	"Bad-Reserved-2Ch",			/* 2Ch */
	"Bad-Reserved-2Dh",			/* 2Dh */
	"Bad-Reserved-2Eh",			/* 2Eh */
	"Bad-Reserved-2Fh",			/* 2Fh */
	"ACA ACTIVE"				/* 30h */
};

static const char *PeripheralDeviceTypeString[32] = {
	"Direct-access",		/* 00h */
	"Sequential-access",		/* 01h */
	"Printer",			/* 02h */
	"Processor",			/* 03h */
	//	"Write-Once-Read-Multiple",	/* 04h */
	"WORM",				/* 04h */
	"CD-ROM",			/* 05h */
	"Scanner",			/* 06h */
	"Optical memory",		/* 07h */
	"Media Changer",		/* 08h */
	"Communications",		/* 09h */
	"(Graphics arts pre-press)",	/* 0Ah */
	"(Graphics arts pre-press)",	/* 0Bh */
	"Array controller",		/* 0Ch */
	"Enclosure services",		/* 0Dh */
	"Simplified direct-access",	/* 0Eh */
	"Reserved-0Fh",			/* 0Fh */
	"Reserved-10h",			/* 10h */
	"Reserved-11h",			/* 11h */
	"Reserved-12h",			/* 12h */
	"Reserved-13h",			/* 13h */
	"Reserved-14h",			/* 14h */
	"Reserved-15h",			/* 15h */
	"Reserved-16h",			/* 16h */
	"Reserved-17h",			/* 17h */
	"Reserved-18h",			/* 18h */
	"Reserved-19h",			/* 19h */
	"Reserved-1Ah",			/* 1Ah */
	"Reserved-1Bh",			/* 1Bh */
	"Reserved-1Ch",			/* 1Ch */
	"Reserved-1Dh",			/* 1Dh */
	"Reserved-1Eh",			/* 1Eh */
	"Unknown"			/* 1Fh */
};

static const char *SenseKeyString[] = {
	"NO SENSE",			/* 0h */
	"RECOVERED ERROR",		/* 1h */
	"NOT READY",			/* 2h */
	"MEDIUM ERROR",			/* 3h */
	"HARDWARE ERROR",		/* 4h */
	"ILLEGAL REQUEST",		/* 5h */
	"UNIT ATTENTION",		/* 6h */
	"DATA PROTECT",			/* 7h */
	"BLANK CHECK",			/* 8h */
	"VENDOR-SPECIFIC",		/* 9h */
	"ABORTED COPY",			/* Ah */
	"ABORTED COMMAND",		/* Bh */
	"EQUAL (obsolete)",		/* Ch */
	"VOLUME OVERFLOW",		/* Dh */
	"MISCOMPARE",			/* Eh */
	"RESERVED",			/* Fh */
};

/*
 *  Ok, can we reduce all of the below to a single sprintf()?
 *---
SCSI Error Report =-=-=-=-=-=-=-=-=-=-=-=-=-= (ioc0,scsi0:0)
  SCSI_Status=02h (CHECK CONDITION)
  Original_CDB[]: 00 00 00 00 00 00 - TestUnitReady
  SenseData[12h]: 70 00 06 00 00 00 00 0A 00 00 00 00 29 00 03 00 00 00
  SenseKey=6h (UNIT ATTENTION); FRU=03h
  ASC/ASCQ=29h/00h, "POWER ON, RESET, OR BUS DEVICE RESET OCCURRED"
 *---
 */

int
dump_cdb(char *foo, unsigned char *cdb)
{
	int i, grpCode, cdbLen;
	int l = 0;

        grpCode = cdb[0] >> 5;
        if (grpCode < 1)
            cdbLen = 6;
        else if (grpCode < 3)
            cdbLen = 10;
        else if (grpCode == 5)
            cdbLen = 12;
        else
            cdbLen = 16;

        for (i=0; i < cdbLen; i++)
            l += sprintf(foo+l, " %02X", cdb[i]);

	return l;
}

int
dump_sd(char *foo, unsigned char *sd)
{
	int snsLen = 8 + SD_Additional_Sense_Length(sd);
	int l = 0;
	int i;

        for (i=0; i < MIN(snsLen,18); i++)
            l += sprintf(foo+l, " %02X", sd[i]);
	l += sprintf(foo+l, "%s", snsLen>18 ? " ..." : "");

	return l;
}

/*  Do ASC/ASCQ lookup to English string  */
char *
ascq_lookup(u8 ASC, u8 ASCQ)
{
	int	 idx;
	int	 specialASC = 0;
	int	 found = 0;
	u8	 savedASCQ = 0;
	char	*ascqStr = NULL;

	if (ASCQ_TablePtr == NULL)
		return NULL;

	/*
	 *  Need to check ASC here; if it is "special," then
	 *  the ASCQ is variable, and indicates failed component number.
	 *  We must treat the ASCQ as a "don't care" while searching the
	 *  ASCQ_Table[] by masking it off, and then restoring it later
	 *  on when we actually need to identify the failed component.
	 */
	if (SPECIAL_ASCQ(ASC,ASCQ))
		ASCQ = 0xFF;

	/*  OK, now search ASCQ_Table[] for a matching entry  */
	for (idx = 0; ASCQ_TablePtr && idx < ASCQ_TableSz; idx++)
		if ((ASC == ASCQ_TablePtr[idx].ASC) && (ASCQ == ASCQ_TablePtr[idx].ASCQ))
			return (ascqStr = ASCQ_TablePtr[idx].Description);

	if ((ASC >= 0x80) || (ASCQ >= 0x80))
		ascqStr = "(Vendor Unique)";
	else
		ascqStr = "(no matching ASC/ASCQ description found)";

	return ascqStr;
}

int
doit(IO_Info_t *ioop, char *foo)
{
	const char	*opstr		= ScsiOpcodeString[ioop->cdbPtr[0]];
	int		 oplen		= strlen(opstr);
	int		 sk		= SD_Sense_Key(ioop->sensePtr);
	const char	*skstr		= SenseKeyString[sk];
	int		 skstrlen 	= strlen(skstr);
	unsigned char	 asc		= SD_ASC(ioop->sensePtr);
	unsigned char	 ascq		= SD_ASCQ(ioop->sensePtr);
	const char	*ascqStr;
	int		 l;

	l = sprintf(foo, "SCSI Error Report =-=-= (%s)\n"
	  "  SCSI_Status=%02Xh (%s)\n"
	  "  Original_CDB[]:",
			ioop->DevIDStr,
			ioop->SCSIStatus,
			ScsiStatusString[ioop->SCSIStatus] );
	l += dump_cdb(foo+l, ioop->cdbPtr);
	l += sprintf(foo+l, " - %s", opstr);
	l += sprintf(foo+l, "\n  SenseData[%02Xh]:", 8+SD_Additional_Sense_Length(ioop->sensePtr));
	l += dump_sd(foo+l, ioop->sensePtr);
	l += sprintf(foo+l, "\n  SenseKey=%Xh (%s); FRU=%02Xh\n  ASC/ASCQ=%02Xh/%02Xh",
			sk, skstr, SD_FRU(ioop->sensePtr), asc, ascq );

	if ((ascqStr = ascq_lookup(asc, ascq)) != NULL) {
		if (ascqStr[0] != '(')
			l += sprintf(foo+l, " \"%s\"", ascqStr);
		else
			l += sprintf(foo+l, " %s", ascqStr);
	}

//	if (SPECIAL_ASCQ(asc,ascq))
//		l += sprintf(foo+l, " (%02Xh)", ascq);

	return l;
}



/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  SCSI_StatusErr_String() - Analyze SCSI Status byte, and sprintf the
 *                            value and English text meaning to msg.
 *                            Plus process NetWare completion code,
 *                            DoRetry, DoDisplay, msg fmting, etc.
 *
 *    NOTE: Assume that msg may not be empty string when passed in.
 *
 *  Return: ptr to msg (passed in).
 */
static char *
SCSI_StatusErr_String(char *msg, u8 SCSIStatus, u32 *ComplPtr, u8 *DoRetry, u8 *pDoDisplay)
{
	int len = strlen(msg);

    /*
     * Figure out why *original* ioop failed and convert the
     * reason to English.  We know it was a CONTINGENT ALLEGIANCE
     * condition; we just need to determine which of three
     * possible SCSI status values caused it.  If it wasn't one
     * of these, then it had to be UNEXPECTED DISCONNECT.
     */
    switch (SCSIStatus & STS_VALID_MASK)
    {
        case STS_CHECK_CONDITION:
            len += sprintf(msg+len, "SCSI_Status=%02Xh (CHECK CONDITION)", SCSIStatus);
            break;
        case STS_CONDITION_MET:
            *DoRetry = FALSE;
//            if (! ScsiDebug)
//                *pDoDisplay = FALSE;
//            *ComplPtr = RecoveredError;
            len += sprintf(msg+len, "SCSI_Status=%02Xh (CONDITION MET)", SCSIStatus);
            break;
        case STS_COMMAND_TERMINATED:
            len += sprintf(msg+len, "SCSI_Status=%02Xh (COMMAND TERMINATED)", SCSIStatus);
            break;
        default:
            len += sprintf(msg+len, "UNEXPECTED DISCONNECT!!!");
    }
    return(msg);
}

#if 0
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  SCSI_Status_String() - Analyze SCSI Status byte, and sprintf the value
 *                         and English text meaning to msg.
 *
 *    NOTE: Assume that msg may not be empty string when passed in.
 *
 *  Return: nothing (+formatted string to msg argument).
 */
static char *
SCSI_Status_String(char *msg, u8 SCSI_Status)
{
	int len = strlen(msg);

    len += sprintf(msg+len, "SCSI_Status=%02Xh ", SCSI_Status);
    switch (SCSI_Status & STS_VALID_MASK)
    {
        case STS_GOOD:
            len += sprintf(msg+len, "(GOOD)");
            break;
        case STS_CHECK_CONDITION:
            len += sprintf(msg+len, "(CHECK CONDITION)");
            break;
        case STS_CONDITION_MET:
            len += sprintf(msg+len, "(CONDITION MET)");
            break;
        case STS_BUSY:
            len += sprintf(msg+len, "(BUSY)");
            break;
        case STS_INTERMEDIATE:
            len += sprintf(msg+len, "(INTERMEDIATE)");
            break;
        case STS_INTERMEDIATE_CONDITION_MET:
            len += sprintf(msg+len, "(INTERMEDIATE-CONDITION MET)");
            break;
        case STS_RESERVATION_CONFLICT:
            len += sprintf(msg+len, "(RESERVATION CONFLICT)");
            break;
        case STS_COMMAND_TERMINATED:
            len += sprintf(msg+len, "(COMMAND TERMINATED)");
            break;
        case STS_QUEUE_FULL:
            len += sprintf(msg+len, "(QUEUE FULL)");
            break;
        default:
            /*
             *  Use backslash(s) to avoid "Trigraph" warning
             */
            len += sprintf(msg+len, "(\?\?)");
            break;
    }
    return(msg);
}
#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  SCSI_OpCode_String() - Analyze SCSI OpCode, and sprintf the value
 *                         and English text meaning to msg.
 *
 *    NOTE: Assume that msg may not be empty string when passed in.
 *
 *  Return: nothing (+formatted string to msg argument).
 */
static char *
SCSI_OpCode_String(char *msg, u8 *CDBPtr, u8 *InqPtr)
{
	switch(CDBPtr[0])
	{
	case CMD_Read10:
		strcat(msg, "READ(10)");		break;
	case CMD_Write10:
		strcat(msg, "WRITE(10)");		break;
	case CMD_TestUnitReady:
		strcat(msg, "TEST UNIT READY");		break;
	case CMD_Inquiry:
		strcat(msg, "INQUIRY");			break;
	case CMD_RequestSense:
		strcat(msg, "REQUEST SENSE");		break;
	case CMD_ReadCapacity:
		strcat(msg, "READ CAPACITY");		break;
	case CMD_Read6:
		strcat(msg, "READ(06)");		break;
	case CMD_Write6:
		strcat(msg, "WRITE(06)");		break;
	case CMD_ModeSense10:
		strcat(msg, "MODE SENSE(10)");		break;
	case CMD_ModeSense6:
		strcat(msg, "MODE SENSE(06)");		break;
	case CMD_ModeSelect10:
		strcat(msg, "MODE SELECT(10)");		break;
	case CMD_ModeSelect6:
		strcat(msg, "MODE SELECT(06)");		break;
	case CMD_Reserve10:
		strcat(msg, "RESERVE(10)");		break;
	case CMD_Reserve6:
		strcat(msg, "RESERVE(06)");		break;
	case CMD_Release10:
		strcat(msg, "RELEASE(10)");		break;
	case CMD_Release6:
		strcat(msg, "RELEASE(06)");		break;
	case CMD_RezeroUnit:
		if (InqPtr && (INQ_Peripheral_Device_Type(InqPtr) == SCSI_TYPE_SAD))
			strcat(msg, "REWIND");
		else
			strcat(msg, "REZERO");
		break;
	case CMD_FormatUnit:
		strcat(msg, "FORMAT UNIT");		break;
	case CMD_ReassignBlock:
		strcat(msg, "REASSIGN BLOCKS");		break;
	case CMD_WriteFilemark:
		strcat(msg, "WRITE FILEMARK");		break;
	case CMD_Space:
		strcat(msg, "SPACE");			break;
	case CMD_Erase:
		strcat(msg, "ERASE");			break;
	case CMD_StartStopUnit:
		if (InqPtr && (INQ_Peripheral_Device_Type(InqPtr) == SCSI_TYPE_SAD)) {
			/* tape */
			strcat(msg, "LOAD UNLOAD");
//			if (CDBPtr[4] & 0x01)
//				strcat(msg, "LOAD");
//			else
//				strcat(msg, "UNLOAD");
		} else {
			/* other (HD, CDROM, etc.) */
			strcat(msg, "STOP START UNIT");
//			if (CDBPtr[4] & 0x01)
//				strcat(msg, "START UNIT");
//			else
//				strcat(msg, "STOP UNIT");
		}
		break;
	case CMD_ReceiveDiagnostic:
		strcat(msg, "RECEIVE DIAGNOSTIC");	break;
	case CMD_SendDiagnostic:
		strcat(msg, "SEND DIAGNOSTIC");		break;
	case CMD_WriteVerify:
		strcat(msg, "WRITE AND VERIFY(10)");	break;
	case CMD_Verify:
		strcat(msg, "VERIFY(10)");		break;
	case CMD_ReadDefectData:
		strcat(msg, "READ DEFECT DATA(10)");	break;
	case CMD_LogSelect:
		strcat(msg, "LOG SELECT");		break;
	case CMD_LogSense:
		strcat(msg, "LOG SENSE");		break;
	case CMD_ReportLuns:
		strcat(msg, "REPORT LUNS");		break;
	default:
		strcat(msg, "\?\?");
	}
	return(msg);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Sense_Key_Handler() - Convert SCSI Sense Key to formatted English text
 *                        in msg string via str operations.
 *
 *    NOTE: Assume that msg may not be empty string when passed in.
 *
 *  Return: nothing (+formatted string to msg argument).
 */
static char *
Sense_Key_Handler(char *msg, IO_Info_t *ioop, u8 *DoRetry)
{
    /*
     *  Intel/StorageExpress RFC.
     *  Don't retry or display errors for SCSI changer device.
     *  950427 -sralston
     */
    if (INQ_Peripheral_Device_Type(ioop->inqPtr) == SCSI_TYPE_CHANGER)
    {
        *DoRetry = FALSE;
        ioop->DoDisplay = FALSE;
    }

    /* Convert Sense Data Sense Key to it's English meaning */
    switch (SD_Sense_Key(ioop->sensePtr))
    {
        case SK_NO_SENSE:
            strcat(msg, "NO SENSE");
            /*
             *    Following code added per Intel/StorageExpress request
             *    that we not log check conditions for tape devices.  Due
             *    to large number of check conditions returned by tape drives,
             *    we were appending to SYS:SYSTEM/SYS$LOG.ERR file at
             *    rapid rate on StorageExpress systems.
             *    950412 -sralston
             */    
            if (INQ_Peripheral_Device_Type(ioop->inqPtr) == SCSI_TYPE_TAPE)
                ioop->DoDisplay = FALSE;
            break;
        case SK_RECOVERED_ERROR:
            strcat(msg, "RECOVERED ERROR");
//            ioop->ComplCode = RecoveredError;
            *DoRetry = FALSE;
            break;
        case SK_NOT_READY:
            strcat(msg, "NOT READY");
            /*
             * Have to do some device specific stuff here due to:
             *   We will get NOT READY SenseKey on CDROMs and TAPEs
             *   when media is not present; we should not perform
             *   retry for this condition.
             *
             * NOTE: Possible/expected ASC/ASCQ values are:
             *    TAPE: 00/00 - NO ADDITIONAL SENSE INFORMATION
             *   CDROM: 3A/00 - MEDIUM NOT PRESENT (sane CDROMs)
             *          0A/00 - ERROR LOG OVERFLOW (Fujitsu ???? CDROM)
             */
            if (INQ_Removable_Medium(ioop->inqPtr))
            {
                if (SD_Additional_Sense_Length(ioop->sensePtr) >= 6)
                {
                    if (    (    (SD_ASC(ioop->sensePtr) == 0x00)
                              && (SD_ASCQ(ioop->sensePtr) == 0x00) )
                         ||
                            (    (SD_ASC(ioop->sensePtr) == 0x3A)
                              && (SD_ASCQ(ioop->sensePtr) == 0x00) )
                         ||
                            /*
                             *  (workaround/fix/hack)
                             *  !#$% Fujitsu (Model ?) CDROM which
                             *  returns (expected) CHECK CONDITION,
                             *  w/ SenseKey==NOT READY, but with
                             *  the following ASC/ASCQ, when media
                             *  is ejected!!!
                             *  Regardless, don't retry it.
                             *  950415 -sralston
                             */
                            (    (SD_ASC(ioop->sensePtr) == 0x0A)
                              && (SD_ASCQ(ioop->sensePtr) == 0x00) )
                       )
                    {
                        *DoRetry = FALSE;
                        ioop->DoDisplay = FALSE;
                    }
                }
            }

            /*
             * NOTE: Possible/expected ASC/ASCQ values are:
             *
             *  We DO NOT want to retry the following:
             *   {any}: 04/00 - LOGICAL UNIT NOT READY, CAUSE NOT REPORTABLE
             *    TAPE: 04/01 - LOGICAL UNIT IS IN PROCESS OF BECOMING READY
             *   {any}: 04/02 - LOGICAL UNIT NOT READY, INITIALIZING COMMAND REQUIRED
             *   {any}: 04/03 - LOGICAL UNIT NOT READY, MANUAL INTERVENTION REQUIRED
             *
             *  We DO want to retry the following, EXCEPT for tapes:
             *    MOST: 04/01 - LOGICAL UNIT IS IN PROCESS OF BECOMING READY
             */
            if (SD_Additional_Sense_Length(ioop->sensePtr) >= 6)
            {
                if (    (SD_ASC(ioop->sensePtr) == 0x04)
                     && (SD_ASCQ(ioop->sensePtr) <= 0x03)
                   )
                {
                    if (    (SD_ASCQ(ioop->sensePtr) != 1)
                         || (INQ_Peripheral_Device_Type(ioop->inqPtr) == SCSI_TYPE_TAPE)
					   )
                    {
                        *DoRetry = FALSE;
                        ioop->DoDisplay = FALSE;
                    }
                }
            }

            break;
        case SK_MEDIUM_ERROR:
            strcat(msg, "MEDIUM ERROR");
//            ioop->ComplCode = MediaError;
            break;
        case SK_HARDWARE_ERROR:
            strcat(msg, "HARDWARE ERROR");
            break;
        case SK_ILLEGAL_REQUEST:
            strcat(msg, "ILLEGAL REQUEST");
//            ioop->ComplCode = DeviceParameterError;
            *DoRetry = FALSE;
#ifndef DEBUG
	    ioop->DoDisplay = FALSE;
#endif
            break;
        case SK_UNIT_ATTENTION:
            strcat(msg, "UNIT ATTENTION");
            /*
             * More device specific stuff here due to:
             * - Yee old ASC/ASCQ of 29h/00h (see below), in which case
             *   we probably will want to do a retry.
             * - Some TAPE devices (Archive 150), will return UA with
             *   with NO ADD'L SENSE, and if retried, then return NOT READY
             *   w/ NO ADD'L SENSE; let code above handle that.  This
             *   happens when no media is present.
             * - Some TAPE devices (HP 4mm DAT) will return UA with
             *   ASC/ASCQ = 00/01 (FILEMARK DETECTED) upon hitting filemark.
             *
             * NOTE: Possible/expected ASC/ASCQ values are:
             *      HD: 29/00 - POWER ON, RESET, OR BUS DEVICE RESET OCCURRED
             *   CDROM: 29/00 - POWER ON, RESET, OR BUS DEVICE RESET OCCURRED
             *      HD: 29/01 - (DEC DSP3210 - POWER ON OCCURED, per SCSI-3)
             *      HD: 29/02 - (DEC DSP3210 - SCSI BUS RESET OCCURED, SCSI-3)
             *      HD: 29/03 - (DEC DSP3210 - BUS DEVICE RESET OCCURED, SCSI-3)
             *    TAPE: 00/00 - NO ADDITIONAL SENSE INFORMATION
             *    TAPE: 00/01 - FILEMARK DETECTED
             *   (any): 28/00 - NOT READY TO READY TRANSITION, MEDIUM MAY
             *                    HAVE CHANGED
             *   (any): 28/01 - IMPORT OR EXPORT ELEMENT ACCESSED
             */
            if (SD_Additional_Sense_Length(ioop->sensePtr) >= 6)
            {
                if (    (SD_ASC(ioop->sensePtr) == 0x29)
                     && (SD_ASCQ(ioop->sensePtr) <= 0x03)
                   )
                {
                    ioop->DoDisplay = FALSE;
                }

                if (    (SD_ASC(ioop->sensePtr) == 0x28)
                     && (    (SD_ASCQ(ioop->sensePtr) == 0x00)
                          || (SD_ASCQ(ioop->sensePtr) == 0x01)
                        )
                   )
                {
                    ioop->DoDisplay = FALSE;
                    *DoRetry = FALSE;
                }

                /* handle tape */
                if (    (INQ_Peripheral_Device_Type(ioop->inqPtr)
                          == SCSI_TYPE_TAPE)
                     &&
                        (    (SD_ASC(ioop->sensePtr) == 0x00)
                          && (    (SD_ASCQ(ioop->sensePtr) == 0x00)
                               || (SD_ASCQ(ioop->sensePtr) == 0x01)
                             )
                        )
                   )
                {
                    ioop->DoDisplay = FALSE;
                    *DoRetry = FALSE;
                }
            }
            break;
        case SK_DATA_PROTECT:
            strcat(msg, "DATA PROTECT");
//            ioop->ComplCode = MediaWriteProtected;
            *DoRetry = FALSE;
            break;
        case SK_BLANK_CHECK:
            strcat(msg, "BLANK CHECK");
            /*
             *  Following code added per Intel/StorageExpress request
             *  that we not log check conditions for tape devices.  Due
             *  to large number of check conditions returned by tape drives,
             *  we were appending to SYS:SYSTEM/SYS$LOG.ERR file at
             *  rapid rate on StorageExpress systems.
             *  950420 -sralston
             *
             * NOTE: Possible/expected ASC/ASCQ values are:
             *    TAPE: 00/01 - END-OF-DATA DETECTED
             */    
            if (INQ_Peripheral_Device_Type(ioop->inqPtr) == SCSI_TYPE_TAPE)
                ioop->DoDisplay = FALSE;
            break;
        case SK_VENDOR_SPECIFIC:
            strcat(msg, "Vendor Specific");
            break;
        case SK_COPY_ABORTED:
            strcat(msg, "COPY ABORTED");
            break;
        case SK_ABORTED_COMMAND:
            strcat(msg, "ABORTED COMMAND");
            break;
        case SK_EQUAL:
            strcat(msg, "EQUAL");
            break;
        case SK_VOLUME_OVERFLOW:
            strcat(msg, "VOLUME OVERFLOW");
            break;
        case SK_MISCOMPARE:
            strcat(msg, "MISCOMPARE");
            break;
        case SK_RESERVED:
            strcat(msg, "RESERVED");
    }
    return(msg);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Sense_Code_Handler() - Convert SCSI Additional Sense Code and
 *               ASC_Qualifier to formatted English text
 *               in msg string via str operations.
 *
 *    NOTE: Assume that msg may not be empty string when passed in.
 *
 *  Return: nothing (+formatted string to msg argument).
 */
void
Sense_Code_Handler(u8 ASC, u8 ASCQ, char *msg)
{
    int    idx;
    int    SpecialASC = FALSE;
    int    Found = FALSE;
    u8     ASCQsave = 0;
	int len;

    /* Convert Additional Sense Code & Qualifier to it's English meaning */

    /*
     * Need to check ASC here; if it is "special" (0x40), then
     * the ASCQ is variable, and indicates failed component number.
     * We must treat the ASCQ as a "don't care" while searching the
     * ASCQ_Table[] by masking it off, and then restoring it later
     * on when we actually need to identify the failed component.
     */
    if ((ASC == 0x40 && ASCQ != 0x00) || (ASC == 0x4D) || (ASC == 0x5D) || (ASC == 0x70))
    {
        SpecialASC = TRUE;
        ASCQsave = ASCQ;
        /* Mask the ASCQ while searching ASCQ_Table[] */
        ASCQ = 0xFF;
    }

    /* OK, now search ASCQ_Table[] for a matching entry */
    for (idx = 0; ASCQ_TablePtr && idx < ASCQ_TableSz; idx++)
    {

        if ((ASC == ASCQ_TablePtr[idx].ASC) && (ASCQ == ASCQ_TablePtr[idx].ASCQ))
        {
            Found = TRUE;
            break;
        }
    }

    /* Restore ASCQ in case the ASC was "special" */
    if (SpecialASC)
        /* Restore ASCQ */
        ASCQ = ASCQsave;

len = strlen(msg);

    if (Found && ASCQ_TablePtr)
    {
        /* add actual error text to msg */
	len += sprintf(msg+len, "\"%s", ASCQ_TablePtr[idx].Description);
        /*
         * For SpecialASC (0x40, 0x70, or 0x4D),
         * tack on the ASCQ hex code at the end of the
         * ASC_ASCQ error text.
         */
        if (SpecialASC)
            len += sprintf(msg+len, " %02Xh)", ASCQ);
        else {
            strcat(msg, "\"");
		len++;
        }
    }
    else
    {
        if (ASCQ_TablePtr == NULL)
            strcat(msg, "(foo)");
        else if ((ASC >= 0x80) || (ASCQ >= 0x80))
            strcat(msg, "(Vendor Unique)");
        else
            strcat(msg, "(no matching ASC/ASCQ description found)");
    }
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Sense_Key_Specific() - If Sense_Key_Specific_Valid bit is set,
 *                         then print additional information via
 *                         a call to SDMS_SystemAlert().
 *
 *  Return: nothing
 */
static void
Sense_Key_Specific(IO_Info_t *ioop, char *msg1)
{
    u8 *sd;
    u8  BadValue;
    u8  SenseKey;
    int    Offset;
	int len = strlen(msg1);

    sd = ioop->sensePtr;
    if (SD_Additional_Sense_Length(sd) < 8)
        return;

    SenseKey = SD_Sense_Key(sd);

    if (SD_Sense_Key_Specific_Valid(sd))
    {
        if (SenseKey == SK_ILLEGAL_REQUEST)
        {
            Offset = SD_Bad_Byte(sd);
            if (SD_Was_Illegal_Request(sd))
            {
                BadValue = ioop->cdbPtr[Offset];
                len += sprintf(msg1+len, "\n  Illegal CDB value=%02Xh found at CDB ",
                    BadValue);
            }
            else
            {
                BadValue = ioop->dataPtr[Offset];
                len += sprintf(msg1+len, "\n  Illegal DATA value=%02Xh found at DATA ",
                    BadValue);
            }
            len += sprintf(msg1+len, "byte=%02Xh", Offset);
            if (SD_SKS_Bit_Pointer_Valid(sd))
                len += sprintf(msg1+len, "/bit=%1Xh", SD_SKS_Bit_Pointer(sd));
        }
        else if (
            (SenseKey == SK_RECOVERED_ERROR) ||
            (SenseKey == SK_HARDWARE_ERROR) ||
            (SenseKey == SK_MEDIUM_ERROR)
          )
        {
            len += sprintf(msg1+len, "\n  Recovery algorithm Actual_Retry_Count=%02Xh",
                SD_Actual_Retry_Count(sd));
        }
    }
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static int
Generic_Sense(IO_Info_t *ioop)
{
    char    msg1[512];
    u8      DoRetry;
    u8      ASC, ASCQ, FRUCode;
    u8     *pDoDisplay;
    u32    *complPtr;
    u32    *pNotifyL;
	int len;
	int addlen;
	u8 snskey;

#if 0
    if (StrISame(REF_STR(s_SYM8XXNW_UNLOADING), ioop->DevIDStr))
    {
        ScsiAlert(0L, 0L, ioop->DevIDStr);
        return 0;
    }
#endif

    /*
     *  Allow caller to specify whether we should display messages.
     *  (TRUE/FALSE, 1/0)  We just set ptr to location here.
     */
    pDoDisplay = &ioop->DoDisplay;

    pNotifyL = &ioop->NotifyL;

    /*
     *  Use OS completion code value passed in,
     *  or use a default value, which may be overridden below.
     */
    if (! ioop->ComplCode)
    {
//        ioop->ComplCode = NonMediaError;
    }
    complPtr = &ioop->ComplCode;

    /*
     *  Set default value for performing a retry of the original IO
     *  to TRUE.  Subject to change during Sense Data analysis.
     *  We shall return this value.
     */
    DoRetry = TRUE;

    /*
     *  Protect ourselves...
     */
    if (ioop->cdbPtr == NULL)
        ioop->cdbPtr = dummyCDB;
    if (ioop->sensePtr == NULL)
        ioop->sensePtr = dummySenseData;
    if (ioop->inqPtr == NULL)
        ioop->inqPtr = dummyInqData;

    /*
     *  Do generic analysis of sense data
     */

        snskey = SD_Sense_Key(ioop->sensePtr);
	addlen = SD_Additional_Sense_Length(ioop->sensePtr);

#if 1		//{

	DoRetry = doit(ioop, msg1);

#else

    /*
     *  Start message with generic SCSI Error Report header + DriveID
     */
    len = sprintf(msg1, "MPI SCSI Error Report =-=-=-=-=-=-=-=-=-=-=-= (%s)\n  ",
                ioop->DevIDStr);

    /*
     *  Analyze SCSI_Status byte of the original ioop that
     *  caused us to perform the AutoReqSense.
     */
    (void) SCSI_StatusErr_String(msg1,
          ioop->SCSIStatus,
          complPtr,
          &DoRetry,
          pDoDisplay);
len = strlen(msg1);

    /*
     *  Always dump CDB //if SenseKey was ILLEGAL_REQUEST
     */
//    if (SenseKey == SK_ILLEGAL_REQUEST)
    {
		int i, grpCode, cdbLen;

        grpCode = ioop->cdbPtr[0] >> 5;
        if (grpCode < 1)
            cdbLen = 6;
        else if (grpCode < 3)
            cdbLen = 10;
        else if (grpCode == 5)
            cdbLen = 12;
        else
            cdbLen = 16;

        len += sprintf(msg1+len, "\n  Original_CDB[]:");
        for (i=0; i < cdbLen; i++)
            len += sprintf(msg1+len, " %02X", ioop->cdbPtr[i]);
        len += sprintf(msg1+len, " - ");
        (void) SCSI_OpCode_String(msg1, ioop->cdbPtr, ioop->inqPtr);
len = strlen(msg1);
        len += sprintf(msg1+len, "\n");
    }

#if 0
    /*
     *  Concatenate the SCSI_OpCode + description
     */
    len += sprintf(msg1+len, "; SCSI_OpCode=%02Xh (", ioop->cdbPtr[0]);
    (void) SCSI_OpCode_String(msg1, ioop->cdbPtr, ioop->inqPtr);
len = strlen(msg1);
    len += sprintf(msg1+len, ")\n");
#endif

// test!
{
	int snsLen = 8;
	int i;

	snsLen += addlen;
	len += sprintf(msg1+len, "  SenseData[%02Xh]:", snsLen);
        for (i=0; i < MIN(snsLen,18); i++)
            len += sprintf(msg1+len, " %02X", ioop->sensePtr[i]);
	len += sprintf(msg1+len, "%s\n", snsLen>18 ? " ..." : "");
}

    /*
     *  Analyze Sense_Key
     *    The following call is powerful/dangerous, because
     *    we allow it to determine whether to perform a retry
     *    and when NOT to display errors for certain conditions.
     */
    len += sprintf(msg1+len, "  SenseKey=%1Xh (",
          SD_Sense_Key(ioop->sensePtr));
    (void) Sense_Key_Handler(msg1, ioop, &DoRetry);
len = strlen(msg1);
    len += sprintf(msg1+len, ")");

    if (
//	    *pDoDisplay &&
            (addlen >= 7)
         && ((FRUCode = SD_Field_Replaceable_Unit_Code(ioop->sensePtr)) != 0)
       )
    {
        /*
         *  Non-zero FRU Code means vendor unique info.
         */
        len += sprintf(msg1+len, "; FRU=%02Xh", FRUCode);
    }

#if 0
    if (ScsiDebug > 4)
    {
        *pDoDisplay = TRUE;
        *pNotifyL = NOTIFY_BOTH;
    }
#endif

//    if (*pDoDisplay)
//    {
////        if (addlen >= 6)
////            strcat(msg1, "\n    ...");
////			len += 8;

        /*
         *  Output multi-line message built so far.
         *  It should look something like this (on the console):
         *

                          *- filled in by NetWare
                         /
vvvvvvvvvvvvvvvvvvvvvvvv/
mm/dd/yy hh:mmp:  0.0.0 (SDMSNET4) DevID=B80x0x0x
(symsense) SCSI Error Report:                              DevID=[V0B8-A1-D6:3]
  SCSI_OpCode=55h (ModeSelect10)
  SCSI_Status=02h (CHECK CONDITION); SenseKey=1h (RECOVERED ERROR)

         *
         * NOTE: NetWare breaks if we attempt to construct
         * a message longer than about 195 bytes.
         */
//        ScsiAlert(*pNotifyL, SEVERITY_RECOVERABLE, "%s", msg1);

        if (addlen >= 6)
        {
            /*
             * Clear out msg1 to start the second message.
             */
//            msg1[0] = '\0';
//            strcat(msg1, "\n    ... Additional sense data information:\n");
////////            strcat(msg1 + strlen(msg1), "\n  Additional sense data information:\n");

            ASC = SD_Additional_Sense_Code(ioop->sensePtr);
            ASCQ = SD_Additional_Sense_Code_Qualifier(ioop->sensePtr);
            /*
             * Analyze ASC/ASCQ
             */
            len += sprintf(msg1+len, "\n  ASC/ASCQ=%02Xh/%02Xh ", ASC, ASCQ);
            Sense_Code_Handler(ASC, ASCQ, msg1);
        }
//    }

    if (
//            *pDoDisplay &&
            (addlen >= 6)
       )
    {
        Sense_Key_Specific(ioop, msg1);
len = strlen(msg1);

        /*
         *  NOTE ! NOTE ! NOTE !
         *  There are still generic items in the std sense data
         *  that need to be checked here, such as:
         *  Command_Specific_Information
         *  (REASSIGN BLOCKS failed block).
         */
    }

#if 0
    if (*pDoDisplay)
    {
        ScsiAlert(*pNotifyL, SEVERITY_RECOVERABLE, "%s", msg1);
    }
#endif

#endif		//}
PrintF(("%s\n", msg1));

    if (DoRetry)
        return(RETRY_STATUS);
    else
        return(PUT_STATUS);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

